/*****************************************************************
 * Copyright (c) ParaSoft Corporation 1991.  All rights reserved.*
 * THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF ParaSoft       *
 * The copyright notice above does not evidence any              *
 * actual or intended publication of such source code.           *
 *****************************************************************/

#include "extern.h"
#include <math.h>

Etype *se_intrin_etype;
Shape *se_intrin_shape;
Typ *se_intrin_typ;

/***********************************************************************************/

static Node *boundary, *boundary_present;
static Node *dim, *dim_present, *stmt;
static int dim_not_present, rank, rank_shift, rank_boundary;
static char var [40], str [80], *setype;
static Etype *etype;
static Etype *etype_dim, *etype_shift;

#define set_r(r)	r = RANK_FROM_SHAPE (sub->Subscript.Expr->com.typ->shape)
#define set_e(e)	e = sub->Subscript.Expr->com.typ->etype
#define set_re(r,e)	{set_r(r); set_e(e);}
/***********************************************************************************/

#define S_NO	1
#define S_O		2
#define A_NO	3
#define A_O		4

#define MAX_NUM_ARGS_INTR	10

static char *arg_name [MAX_NUM_ARGS_INTR];
static int arg_type [MAX_NUM_ARGS_INTR];
static int arg_num;

static int arg_rank [MAX_NUM_ARGS_INTR];
static int arg_present [MAX_NUM_ARGS_INTR];
static Node *arg_node [MAX_NUM_ARGS_INTR];
static Etype *arg_etype [MAX_NUM_ARGS_INTR];

static is_present (sublist)
Node *sublist;
{
	if (strcmp (sublist->lst.value->Subscript.Expr->com.str, ".FALSE."))
		return 1;
	else
		return 0;
}

static fill_args (sublist)
Node *sublist;
{
	Node *sub;
	int i;

	if (sublist == NN)
		return 1;

	for (i = 0; i < arg_num; i++) {
		sub = sublist->lst.value;

		switch (arg_type [i]) {

		case S_NO:
			arg_rank [i] = RANK_FROM_SHAPE (sub->Subscript.Expr->com.typ->shape);
		
                        /* BOZKUS */
            /* this MASK if mask rank 0 it means NO mask especially SUM */
                        if ( arg_rank[i]!=0)            
                         	arg_present [i] = 1;
			arg_node [i] = sub->Subscript.Expr;
			arg_etype [i] = sub->Subscript.Expr->com.typ->etype;
			break;

		case S_O:
			arg_rank [i] = RANK_FROM_SHAPE (sub->Subscript.Expr->com.typ->shape);
			arg_node [i] = sub->Subscript.Expr;
			arg_etype [i] = sub->Subscript.Expr->com.typ->etype;
			sublist = sublist->lst.next;
			arg_present [i] = is_present (sublist);
			break;

		case A_NO:
			arg_rank [i] = RANK_FROM_SHAPE (sub->Subscript.Expr->com.typ->shape);
			arg_present [i] = 1;
			arg_node [i] = sub->Subscript.Expr;
			arg_etype [i] = sub->Subscript.Expr->com.typ->etype;
			sublist = sublist->lst.next;
			break;

		case A_O:
			arg_rank [i] = RANK_FROM_SHAPE (sub->Subscript.Expr->com.typ->shape);
			arg_node [i] = sub->Subscript.Expr;
			arg_etype [i] = sub->Subscript.Expr->com.typ->etype;
			sublist = sublist->lst.next;
			arg_present [i] = is_present (sublist);
			sublist = sublist->lst.next;
			break;

#ifdef CHECK
		default:
			SYSERR;
#endif
		}

		sublist = sublist->lst.next;
	}

	return 0;
}

/***********************************************************************************/

static char *proc;

static must_not_be_lcx (a)
int a;
{
	int type = arg_etype [a]->type;

	if (type == TYP_LOGICAL
	||  type == TYP_CHARACTER
	||  type == TYP_COMPLEX) {
		sprintf (serr, "argument %s of %s cannot be %s"
		, arg_name [a], proc, etype_name (arg_etype [a]));
		fatal (serr);
	}
}

static must_not_be_lc (a)
int a;
{
	int type = arg_etype [a]->type;

	if (type == TYP_LOGICAL
	||  type == TYP_CHARACTER) {
		sprintf (serr, "argument %s of %s cannot be %s"
		, arg_name [a], proc, etype_name (arg_etype [a]));
		fatal (serr);
	}
}

static must_not_be_c (a)
int a;
{
	int type = arg_etype [a]->type;

	if (type == TYP_CHARACTER) {
		sprintf (serr, "argument %s of %s cannot be %s"
		, arg_name [a], proc, etype_name (arg_etype [a]));
		fatal (serr);
	}
}

static must_be_conformal_to (a1, a2)
int a1, a2;
{
	if (arg_present [a1] && arg_present [a2]
	&&  arg_rank [a2] && arg_rank [a2] != arg_rank [a1]) {
		sprintf (serr, "argument %s of %s must be conformal with argument %s"
		, arg_name [a2], proc, arg_name [a1]);
		fatal (serr);
	}
}

static must_be_of_the_same_shape (a1, a2)
int a1, a2;
{
    if (arg_present [a1] && arg_present [a2]
    && arg_rank [a2] != arg_rank [a1]) {
        sprintf (serr, "argument %s of %s must have the same shape as argument %s"
        , arg_name [a2], proc, arg_name [a1]);
        fatal (serr);
    }
}

static must_be_scalar (a)
int a;
{
	if (arg_rank [a]) {
		sprintf (serr, "argument %s of %s must be scalar, is of rank %d"
		, arg_name [a], proc, arg_rank [a]);
		fatal (serr);
	}
}

static must_be_of_rank_1 (a)
int a;
{
	if (arg_rank [a] != 1) {
		sprintf (serr, "argument %s of %s must be of rank 1, is of rank %d"
		, arg_name [a], proc, arg_rank [a]);
		fatal (serr);
	}
}

static must_be_of_rank_x (a, rank)
int a, rank;
{
	if (arg_rank [a] != rank) {
		sprintf (serr, "argument %s of %s must be of rank %d, is of rank %d"
		, arg_name [a], proc, rank, arg_rank [a]);
		fatal (serr);
	}
}

static must_be_logical (a)
int a;
{
	Etype *etype = arg_etype [a];

	if (etype->type != TYP_LOGICAL) {
		sprintf (serr, "argument %s of %s must be LOGICAL, is %s"
		, arg_name [a], proc, etype_name (etype));
		fatal (serr);
	}
}

static must_be_integer (a)
int a;
{
	Etype *etype = arg_etype [a];

	if (etype->type != TYP_INTEGER) {
		sprintf (serr, "argument %s of %s must be INTEGER, is %s"
		, arg_name [a], proc, etype_name (etype));
		fatal (serr);
	}
}

static check_dim (expr, from, to)
Node *expr;
int from, to;
{
	double value;

	if (expr_is_constant (expr, &value)) {
		int dim = U_nint (value);

		if (dim < from || dim > to) {
			sprintf (serr, "argument DIM of %s should be in the range %d <= DIM <= %d, is %d"
			, proc, from, to, dim);
			fatal (serr);
		}
	}
	else {
		if (flag_cmf)
			warning ("argument DIM is not a constant expression");

/*
		if (expr->com.type == NAME) {
			T_Entry *entry = lookup_all (expr->com.str);

			if (entry != TN && IS_optional (entry))
				fatal ("actual argument DIM cannot be an optional dummy argument");
		}
*/
	}
}

static check_dim_no_cmf_constant (expr, from, to)
Node *expr;
int from, to;
{
	double value;

	if (expr_is_constant (expr, &value)) {
		int dim = U_nint (value);

		if (dim < from || dim > to) {
			sprintf (serr, "argument DIM of %s should be in the range %d <= DIM <= %d, is %d"
			, proc, from, to, dim);
			fatal (serr);
		}
	}
	else {
/*
		if (expr->com.type == NAME) {
			T_Entry *entry = lookup_all (expr->com.str);

			if (entry != TN && IS_optional (entry))
				fatal ("actual argument DIM cannot be an optional dummy argument");
		}
*/
	}
}

must_be_of_the_same_type (a1, a2)
int a1, a2;
{
	if (arg_present [a1] && arg_present [a2]
	&&  etype_ne (arg_etype [a1], arg_etype [a2])) {
		sprintf (serr, "argument %s of %s must be %s, is %s", arg_name [a2], proc
		, etype_name (arg_etype [a1]), etype_name (arg_etype [a2]));
		fatal (serr);
	}
}

/***********************************************************************************/

intrin_present (root)
Node **root;
{
	Node *sublist = (*root)->FunctionReference.SubscriptList;
	Node *sub, *temp;
	char *name = sublist->lst.value->Subscript.Expr->com.str;

	if (name == (char*) 0) {
		fatal ("wrong argument to intrinsic function 'present'");
		*root = NN;
	}
	else
		*root = make_Name (opt_name (name));

	(*root)->com.typ = typ_malloc ();
	(*root)->com.typ->etype = Etype_LOGICAL_0;

	return 0;
}

intrin_a_rank (root)  /*!!! CMF exten */
Node **root;
{
	Node *sublist = (*root)->FunctionReference.SubscriptList;
	Node *sub, *temp;
	int num;
	T_Entry *entry;
	Shape *shape;

	char *name;

	name = sublist->lst.value->Subscript.Expr->com.str;

    {
		Node *list, *expr;

		expr = make_DumExpr_1_PIN (make_IntegerConstant (1));
		list = make_SubscriptList (expr);
		*root = make_ArraySection (make_Name (desc_name (name)), list);
	}

	(*root)->com.typ = typ_malloc ();
	(*root)->com.typ->etype = Etype_INTEGER_0;

	return 0;
}

intrin_rank (root)  /*!!! CMF exten */
Node **root;
{
	Node *sublist = (*root)->FunctionReference.SubscriptList;
	Node *sub, *temp;
	int num;
	T_Entry *entry;
	Shape *shape = sublist->lst.value->Subscript.Expr->com.typ->shape;

/*
	char *name;

	name = sublist->lst.value->Subscript.Expr->com.str;

	if (name != CN && (
	    (entry = lookup_loc (name)) == TN
	||  (shape = entry->typ->shape) == SN))
		*root = make_IntegerConstant (0);
	else
*/
		*root = make_IntegerConstant (RANK_FROM_SHAPE (shape));

	(*root)->com.typ = typ_malloc ();
	(*root)->com.typ->etype = Etype_INTEGER_0;

	return 0;
}

/* There are 3 levels at which they can be evaluated, here we implement the second */

intrin_size (root)
Node **root;
{
	Node *sublist = (*root)->FunctionReference.SubscriptList;
	Node *sub, *temp;
	int num;
	T_Entry *entry;

	char *name;
	Node *dim, *dim_present;

	num = 0;
	for (LIST_LOOP (sub, sublist, temp)) {
		num++;

		switch (num) {
		case 1:
			name = sub->Subscript.Expr->com.str;
			break;

		case 2:
			break;

		case 3:
			dim = sub->Subscript.Expr;
			break;
        }
    }

    {
		Node *list, *expr;

		expr = make_DumExpr_PIN (make_IntegerConstant (1), &Plus, dim);
		list = make_SubscriptList (expr);
		*root = make_ArraySection (make_Name (desc_name (name)), list);
	}

	(*root)->com.typ = typ_malloc ();
	(*root)->com.typ->etype = Etype_INTEGER_0;

	return 0;
}

intrin_a_lbound (root)
Node **root;
{
	Node *sublist = (*root)->FunctionReference.SubscriptList;
	Node *sub, *temp;
	int num;
	T_Entry *entry;

	char *name;
	Node *dim;

	num = 0;
	for (LIST_LOOP (sub, sublist, temp)) {
		num++;

		switch (num) {
		case 1:
			name = sub->Subscript.Expr->com.str;
			break;

		case 2:
			break;

		case 3:
			dim = sub->Subscript.Expr;
			break;
		}
	}

	{
		Node *list, *expr;

		expr = make_DumExpr_PIN (make_IntegerConstant (8), &Plus, dim);
		list = make_SubscriptList (expr);
		*root = make_ArraySection (make_Name (desc_name (name)), list);
	}

	(*root)->com.typ = typ_malloc ();
	(*root)->com.typ->etype = Etype_INTEGER_0;

	return 0;
}

intrin_a_ubound (root)
Node **root;
{
	Node *sublist = (*root)->FunctionReference.SubscriptList;
	Node *sub, *temp;
	int num;
	T_Entry *entry;

	char *name;
	Node *dim;

	num = 0;
	for (LIST_LOOP (sub, sublist, temp)) {
		num++;

		switch (num) {
		case 1:
			name = sub->Subscript.Expr->com.str;
			break;

		case 2:
			break;

		case 3:
			dim = sub->Subscript.Expr;
			break;
		}
	}

	{
		Node *list, *expr;

		expr = make_DumExpr_PIN (make_IntegerConstant (15), &Plus, dim);
		list = make_SubscriptList (expr);
		*root = make_ArraySection (make_Name (desc_name (name)), list);
	}

	(*root)->com.typ = typ_malloc ();
	(*root)->com.typ->etype = Etype_INTEGER_0;

	return 0;
}

intrin_lbound_CMF (root)
Node **root;
{
	proc = "DLBOUND";
	stand ("intrinsic DLBOUND (is called LBOUND in Fortran 90)");
	return intrin_lbound (root);
}

intrin_lbound_F90 (root)
Node **root;
{
	proc = "LBOUND";
	return intrin_lbound (root);
}

intrin_shape_s_F90 (root)
Node **root;
{
	proc = "SHAPE";
	return intrin_shape_s (root);
}

intrin_shape_a_F90 (root)
Node **root;
{
	proc = "SHAPE";
	return intrin_shape_a (root);
}

intrin_shape_s_CMF (root)
Node **root;
{
	proc = "DSHAPE";
	stand ("intrinsic DSHAPE (is called SHAPE in Fortran 90)");
	return intrin_shape_s (root);
}

intrin_shape_a_CMF (root)
Node **root;
{
	proc = "DSHAPE";
	stand ("intrinsic DSHAPE (is called SHAPE in Fortran 90)");
	return intrin_shape_a (root);
}

intrin_shape_s (root)
Node **root;
{
	Node *sublist = (*root)->FunctionReference.SubscriptList;

	arg_num = 1;
	arg_name [0] = "SOURCE"; arg_type [0] = S_NO;

	if (fill_args (sublist))
		return 0;

	return 2;
}

intrin_shape_a (root)
Node **root;
{
	Node *sublist = (*root)->FunctionReference.SubscriptList;

	arg_num = 1;
	arg_name [0] = "SOURCE"; arg_type [0] = S_NO;

	if (fill_args (sublist))
		return 0;

	return 2;
}

intrin_size_F90 (root)
Node **root;
{
	proc = "SIZE";
	return intrin_size_run (root);
}

intrin_size_CMF (root)
Node **root;
{
	proc = "DSIZE";
	stand ("intrinsic DSIZE (is called SIZE in Fortran 90)");
	return intrin_size_run (root);
}

intrin_size_run (root)
Node **root;
{
	Node *sublist = (*root)->FunctionReference.SubscriptList;

	arg_num = 2;
	arg_name [0] = "ARRAY"; arg_type [0] = A_NO;
	arg_name [1] = "DIM";   arg_type [1] = S_O;

	if (fill_args (sublist))
		return 0;

	rank = arg_rank [0];

	if (arg_present [1])
		if (must_be_integer (1) && rank)
			check_dim_no_cmf_constant (arg_node [1], 1, rank);

	if (flag_cmf) {
		(*root)->com.str = "@dsize_fun";
		decl_intrin_add (mac_etype_name (Etype_INTEGER_0), "@dsize_fun");
	}
	else {
		(*root)->com.str = "@size_fun";
		decl_intrin_add (mac_etype_name (Etype_INTEGER_0), "@size_fun");
	}

	return 2;
}

intrin_lbound (root)
Node **root;
{
	Node *sublist = (*root)->FunctionReference.SubscriptList;

	arg_num = 2;
	arg_name [0] = "ARRAY"; arg_type [0] = A_NO;
	arg_name [1] = "DIM";   arg_type [1] = S_O;

	if (fill_args (sublist))
		return 0;

	rank = arg_rank [0];

	if (arg_present [1]) {
		(*root)->com.str = "lbound_s";

		if (must_be_integer (1) && rank)
			check_dim_no_cmf_constant (arg_node [1], 1, rank);
	}
	else
		(*root)->com.str = "lbound_a";

	return 2;
}

intrin_ubound_F90 (root)
Node **root;
{
	proc = "UBOUND";
	return intrin_ubound (root);
}

intrin_ubound_CMF (root)
Node **root;
{
	proc = "DUBOUND";
	stand ("intrinsic DUBOUND (is called UBOUND in Fortran 90)");
	return intrin_ubound (root);
}

intrin_ubound (root)
Node **root;
{
	Node *sublist = (*root)->FunctionReference.SubscriptList;

	arg_num = 2;
	arg_name [0] = "ARRAY"; arg_type [0] = A_NO;
	arg_name [1] = "DIM";   arg_type [1] = S_O;

	if (fill_args (sublist))
		return 0;

	rank = arg_rank [0];

	if (arg_present [1]) {
		(*root)->com.str = "ubound_s";

		decl_intrin_add (mac_etype_name (Etype_INTEGER_0), "ubound_s");

		if (must_be_integer (1) && rank)
			check_dim_no_cmf_constant (arg_node [1], 1, rank);
	}
	else
		(*root)->com.str = "ubound_a";

	return 2;
}

intrin_conjg (root)
Node **root;
{
	Node *sublist = (*root)->FunctionReference.SubscriptList;
	Node *sub, *temp;
	int num;
	T_Entry *entry;

	Node *sublist2;
	Etype *etype = EN;

	num = 0;
	for (LIST_LOOP (sub, sublist, temp)) {
		num++;

		switch (num) {
		case 1:
			etype = sub->Subscript.Expr->com.typ->etype;
			break;
		}
	}

	if (etype->kind == 1)
		(*root)->com.str = "dconjg";

	return 2;
}

intrin_aimag (root)
Node **root;
{
	Node *sublist = (*root)->FunctionReference.SubscriptList;
	Node *sub, *temp;
	int num;
	T_Entry *entry;

	Node *sublist2;
	Etype *etype = EN;

	num = 0;
	for (LIST_LOOP (sub, sublist, temp)) {
		num++;

		switch (num) {
		case 1:
			etype = sub->Subscript.Expr->com.typ->etype;
			break;
		}
	}

	if (etype->kind == 1)
		(*root)->com.str = "dimag";

	return 2;
}

intrin_iand (root)
Node **root;
{
	(*root)->com.str = "and";

	return 2;
}

intrin_ieor(root)
Node **root;
{
	(*root)->com.str = "xor";

	return 2;
}

intrin_ior (root)
Node **root;
{
	(*root)->com.str = "or";

	return 2;
}

intrin_cmplx (root)
Node **root;
{
	Node *sublist = (*root)->FunctionReference.SubscriptList;
	Node *sub, *temp;
	int num;
	T_Entry *entry;

	Node *sublist2, *sublist1, *opt_present;
	Etype *etype = EN;

	num = 0;
	for (LIST_LOOP (sub, sublist, temp)) {
		num++;

		switch (num) {
		case 1:
			sublist1 = temp;
			break;

		case 2:
			sublist2 = temp;
			break;

		case 3:
			opt_present = sub->Subscript.Expr;
			break;

		case 4:
			etype = sub->Subscript.Expr->com.typ->etype;
			break;

		case 5:
			break;
		}
	}

	if (!strncmp (opt_present->com.str, ".FALSE.", 7))
		sublist1->lst.next = NN;
	else
		sublist2->lst.next = NN;

/*!!! the third argument is just ignored */

	if (etype != EN)
		if (etype->kind == 1)
			(*root)->com.str = "dcmplx";  /*!!! OK for sun */

	return 2;
}

intrin_firstloc (root)
Node **root;
{
	Node *sublist = (*root)->FunctionReference.SubscriptList;

	stand ("array intrinsic FIRSTLOC");

	proc = "FIRSTLOC";
	arg_num = 2;
	arg_name [0] = "MASK"; arg_type [0] = A_NO;
	arg_name [1] = "DIM"; arg_type [1] = S_O;
	
	if (fill_args (sublist))
		return 0;

	rank = arg_rank [0];

	if (!arg_present [1])
		dim_not_present = 1;
	else {
		if (must_be_integer (1) && rank)  /* avoids spurious errors for rank == 0 */
			check_dim (arg_node [1], 1, rank);

		dim_not_present = 0;
	}

	if (dim_not_present || rank <= 1) {
		sprintf (var, "firstloc_nd_%d", rank);
		(*root)->com.str = U_str_copy (var);
	}
	else {
		sprintf (var, "firstloc_s_%d", rank);
		(*root)->com.str = U_str_copy (var);
	}

	se_intrin_etype = Etype_LOGICAL_0;
	se_intrin_shape = Shape_a [rank];

	se_intrin_typ = typ_malloc ();
	se_intrin_typ->etype = se_intrin_etype;
	se_intrin_typ->shape = se_intrin_shape;

	return 3;
}

intrin_lastloc (root)
Node **root;
{
	Node *sublist = (*root)->FunctionReference.SubscriptList;

	stand ("array intrinsic LASTLOC");

	proc = "LASTLOC";
	arg_num = 2;
	arg_name [0] = "MASK"; arg_type [0] = A_NO;
	arg_name [1] = "DIM"; arg_type [1] = S_O;
	
	if (fill_args (sublist))
		return 0;

	rank = arg_rank [0];

	if (!arg_present [1])
		dim_not_present = 1;
	else {
		if (must_be_integer (1) && rank)  /* avoids spurious errors for rank == 0 */
			check_dim (arg_node [1], 1, rank);

		dim_not_present = 0;
	}

	if (dim_not_present || rank <= 1) {
		sprintf (var, "lastloc_nd_%d", rank);
		(*root)->com.str = U_str_copy (var);
	}
	else {
		sprintf (var, "lastloc_s_%d", rank);
		(*root)->com.str = U_str_copy (var);
	}

	se_intrin_etype = Etype_LOGICAL_0;
	se_intrin_shape = Shape_a [rank];

	se_intrin_typ = typ_malloc ();
	se_intrin_typ->etype = se_intrin_etype;
	se_intrin_typ->shape = se_intrin_shape;

	return 3;
}

intrin_all (root)
Node **root;
{
	Node *sublist = (*root)->FunctionReference.SubscriptList;

	proc = "ALL";
	arg_num = 2;
	arg_name [0] = "MASK"; arg_type [0] = A_NO;
	arg_name [1] = "DIM"; arg_type [1] = S_O;
	
	if (fill_args (sublist))
		return 0;

	rank = arg_rank [0];

	if (!arg_present [1])
		dim_not_present = 1;
	else {
		if (must_be_integer (1) && rank)  /* avoids spurious errors for rank == 0 */
			check_dim (arg_node [1], 1, rank);

		dim_not_present = 0;
	}

	if (dim_not_present && rank <= 1) {
		sprintf (var, "all_%d_l", rank);
		(*root)->com.str = U_str_copy (var);

		decl_intrin_add (mac_etype_name (Etype_LOGICAL_0), (*root)->com.str);

		se_intrin_shape = Shape_a [0];
	}
	else {
		sprintf (var, "all_%d_l", rank);
                if(dim_not_present) strcat(var,"_s");
                else strcat(var,"_v");
		(*root)->com.str = U_str_copy (var);

		se_intrin_shape = Shape_a [rank - 1];
	}

	se_intrin_etype = Etype_LOGICAL_0;

	se_intrin_typ = typ_malloc ();
	se_intrin_typ->etype = se_intrin_etype;
	se_intrin_typ->shape = se_intrin_shape;

	return 3;
}

intrin_any (root)
Node **root;
{
	Node *sublist = (*root)->FunctionReference.SubscriptList;

	proc = "ANY";
	arg_num = 2;
	arg_name [0] = "MASK"; arg_type [0] = A_NO;
	arg_name [1] = "DIM"; arg_type [1] = S_O;
	
	if (fill_args (sublist))
		return 0;

	rank = arg_rank [0];

	if (!arg_present [1])
		dim_not_present = 1;
	else {
		if (must_be_integer (1) && rank)
			check_dim (arg_node [1], 1, rank);

		dim_not_present = 0;
	}

	if (dim_not_present && rank <= 1) {
		sprintf (var, "any_%d_l", rank);
		(*root)->com.str = U_str_copy (var);

		decl_intrin_add (mac_etype_name (Etype_LOGICAL_0), (*root)->com.str);

		se_intrin_shape = Shape_a [0];
	}
	else {
		sprintf (var, "any_%d_l", rank);
                if(dim_not_present) strcat(var,"_s");
                else strcat(var,"_v");
		(*root)->com.str = U_str_copy (var);

		se_intrin_shape = Shape_a [rank - 1];
	}

	se_intrin_etype = Etype_LOGICAL_0;

	se_intrin_typ = typ_malloc ();
	se_intrin_typ->etype = se_intrin_etype;
	se_intrin_typ->shape = se_intrin_shape;

	return 3;
}

intrin_count (root)
Node **root;
{
	Node *sublist = (*root)->FunctionReference.SubscriptList;

	proc = "COUNT";
	arg_num = 2;
	arg_name [0] = "MASK"; arg_type [0] = A_NO;
	arg_name [1] = "DIM"; arg_type [1] = S_O;
	
	if (fill_args (sublist))
		return 0;

	rank = arg_rank [0];

	if (!arg_present [1])
		dim_not_present = 1;
	else {
		if (must_be_integer (1) && rank)
			check_dim (arg_node [1], 1, rank);

		dim_not_present = 0;
	}

	if (dim_not_present && rank <= 1) {
		sprintf (var, "count_%d_l", rank);
		(*root)->com.str = U_str_copy (var);

		decl_intrin_add (mac_etype_name (Etype_INTEGER_0), (*root)->com.str);

		se_intrin_shape = Shape_a [0];
	}
	else {
		sprintf (var, "count_%d_l", rank);
                if(dim_not_present) strcat(var,"_s");
                else strcat(var,"_v");
		(*root)->com.str = U_str_copy (var);

		se_intrin_shape = Shape_a [rank - 1];
	}

	se_intrin_etype = Etype_INTEGER_0;

	se_intrin_typ = typ_malloc ();
	se_intrin_typ->etype = se_intrin_etype;
	se_intrin_typ->shape = se_intrin_shape;

	return 3;
}

intrin_matmul (root)
Node **root;
{
	Node *sublist = (*root)->FunctionReference.SubscriptList;
	Etype *etype1, *etype2;
	int subtype = 0;
	char *conv;
	int type0, type1, rank1, rank2;
	int num;
	Node *temp, *sub;

	proc = "MATMUL";
	arg_num = 2;
	arg_name [0] = "MATRIX_A"; arg_type [0] = A_NO;
	arg_name [1] = "MATRIX_B"; arg_type [1] = A_NO;

	if (fill_args (sublist))
		return 0;

	type0 = (etype1 = arg_etype [0])->type;
	type1 = (etype2 = arg_etype [1])->type;
	rank1 = arg_rank [0];
	rank2 = arg_rank [1];

	if (type0 == TYP_LOGICAL) {
		if (type1 != TYP_LOGICAL) {
			sprintf (serr, "argument MATRIX_B of %s must be LOGICAL, is %s"
			, proc, etype_name (arg_etype [1]));
			fatal (serr);
		}
	}
	else {
		if (type1 == TYP_LOGICAL) {
			sprintf (serr, "argument MATRIX_B of %s must be of numeric type, is %s"
			, proc, etype_name (arg_etype [1]));
			fatal (serr);
		}
	}

	if (rank1 > 2 || rank2 > 2 || (rank1 == 1 && rank2 == 1)) {
		sprintf (serr, "invalid ranks for the arguments of MATMUL, the ranks are %d and %d"
		, rank1, rank2);
		fatal (serr);
	}

	if (etype1->type != TYP_LOGICAL
	&&  etype2->type != TYP_LOGICAL
	&&  etype_ne (etype1, etype2)) {
		if (etype1->type == etype2->type) {
			if (etype1->kind > etype2->kind) {
				etype = etype1;
				subtype = 1;
			}
			else {
				etype = etype2;
				subtype = 2;
			}
		}
		else {
			if (etype1->type > etype2->type) {
				etype = etype1;
				subtype = 1;
			}
			else {
				etype = etype2;
				subtype = 2;
			}
		}
	}
	else
		etype = etype1;

	if (subtype) {
		switch (etype->type) {

		case TYP_REAL:
			switch (etype->kind) {
			case 0:
				conv = "real";
				break;

			case 1:
				conv = "dble";
				break;
			}
			break;

		case TYP_COMPLEX:
			switch (etype->kind) {
			case 0:
				conv = "cmplx";
				break;

			case 1:
				conv = "dcmplx";
				break;
			}
			break;

#ifdef CHECK
		default:
			SYSERR;
#endif
		}
		
		num = 0;
		for (LIST_LOOP (sub, sublist, temp)) {
			num++;

			switch (num) {
			case 1:
				if (subtype == 1)
					break;

				sub->Subscript.Expr = make_ArraySection (make_Name (conv)
				, make_SubscriptList (sub->Subscript.Expr));
				sub->Subscript.Expr = parse_ArraySection (sub->Subscript.Expr);
				sub->Subscript.Expr = temp_replace (sub->Subscript.Expr);

				break;

			case 3:
				if (subtype == 2)
					break;

				sub->Subscript.Expr = make_ArraySection (make_Name (conv)
				, make_SubscriptList (sub->Subscript.Expr));
				sub->Subscript.Expr = parse_ArraySection (sub->Subscript.Expr);
				sub->Subscript.Expr = temp_replace (sub->Subscript.Expr);

				break;
			}
		}
	}

	setype = mac_etype_short_name (etype);
	
	sprintf (var, "matmul_%d_%s", rank1, setype);
	(*root)->com.str = U_str_copy (var);

	if (rank1 == 2 && rank2 == 2)
		se_intrin_shape = Shape_a [2];
	else
		se_intrin_shape = Shape_a [1];

	se_intrin_etype = etype;

	se_intrin_typ = typ_malloc ();
	se_intrin_typ->etype = se_intrin_etype;
	se_intrin_typ->shape = se_intrin_shape;

	return 3;
}

intrin_dotproduct_CMF (root)
Node **root;
{
	proc = "DOTPRODUCT";
	stand ("intrinsic DOTPRODUCT (is called DOT_PRODUCT in Fortran 90)");
	return intrin_dot_product (root);
}

intrin_dot_product_F90 (root)
Node **root;
{
	proc = "DOT_PRODUCT";
	return intrin_dot_product (root);
}

intrin_dot_product (root)
Node **root;
{
	Node *sublist = (*root)->FunctionReference.SubscriptList;
	Etype *etype1, *etype2;
	int subtype = 0;
	char *conv;
	int type0, type1, rank1, rank2;
	int num;
	Node *temp, *sub;

	proc = "DOT_PRODUCT";
	arg_num = 2;
	arg_name [0] = "VECTOR_A"; arg_type [0] = A_NO;
	arg_name [1] = "VECTOR_B"; arg_type [1] = A_NO;

	if (fill_args (sublist))
		return 0;

    if (arg_rank [0])
        must_be_of_rank_1 (0);

    if (arg_rank [1])
        must_be_of_rank_1 (1);

	type0 = (etype1 = arg_etype [0])->type;
	type1 = (etype2 = arg_etype [1])->type;
	rank1 = arg_rank [0];
	rank2 = arg_rank [1];

	if (type0 == TYP_LOGICAL) {
		if (type1 != TYP_LOGICAL) {
			sprintf (serr, "argument VECTOR_B of %s must be LOGICAL, is %s"
			, proc, etype_name (arg_etype [1]));
			fatal (serr);
		}
	}
	else {
		if (type1 == TYP_LOGICAL) {
			sprintf (serr, "argument VECTOR__B of %s must be of numeric type, is %s"
			, proc, etype_name (arg_etype [1]));
			fatal (serr);
		}
	}

	if (etype1->type != TYP_LOGICAL
	&&  etype2->type != TYP_LOGICAL
	&&  etype_ne (etype1, etype2)) {
		if (etype1->type == etype2->type) {
			if (etype1->kind > etype2->kind) {
				etype = etype1;
				subtype = 1;
			}
			else {
				etype = etype2;
				subtype = 2;
			}
		}
		else {
			if (etype1->type > etype2->type) {
				etype = etype1;
				subtype = 1;
			}
			else {
				etype = etype2;
				subtype = 2;
			}
		}
	}
	else
		etype = etype1;

	if (subtype) {
		switch (etype->type) {

		case TYP_REAL:
			switch (etype->kind) {
			case 0:
				conv = "real";
				break;

			case 1:
				conv = "dble";
				break;
			}
			break;

		case TYP_COMPLEX:
			switch (etype->kind) {
			case 0:
				conv = "cmplx";
				break;

			case 1:
				conv = "dcmplx";
				break;
			}
			break;

#ifdef CHECK
		default:
			SYSERR;
#endif
		}
		
		num = 0;
		for (LIST_LOOP (sub, sublist, temp)) {
			num++;

			switch (num) {
			case 1:
				if (subtype == 1)
					break;

				sub->Subscript.Expr = make_ArraySection (make_Name (conv)
				, make_SubscriptList (sub->Subscript.Expr));
				sub->Subscript.Expr = parse_ArraySection (sub->Subscript.Expr);
				sub->Subscript.Expr = temp_replace (sub->Subscript.Expr);

				break;

			case 3:
				if (subtype == 2)
					break;

				sub->Subscript.Expr = make_ArraySection (make_Name (conv)
				, make_SubscriptList (sub->Subscript.Expr));
				sub->Subscript.Expr = parse_ArraySection (sub->Subscript.Expr);
				sub->Subscript.Expr = temp_replace (sub->Subscript.Expr);

				break;
			}
		}
	}

	setype = mac_etype_short_name (etype);
	
	sprintf (var, "dot_product_1_%s", setype);
	(*root)->com.str = U_str_copy (var);

    decl_intrin_add (mac_etype_name (etype), (*root)->com.str);

	se_intrin_shape = Shape_a [0];

	se_intrin_etype = etype;

	se_intrin_typ = typ_malloc ();
	se_intrin_typ->etype = se_intrin_etype;
	se_intrin_typ->shape = se_intrin_shape;

	return 3;
}

intrin_transpose (root)
Node **root;
{
	Node *sublist = (*root)->FunctionReference.SubscriptList;
	Node *sub, *temp;
	int num;
	T_Entry *entry;

	Node *dim_present, *stmt;
	int dim_not_present, rank;
	char var [20], str [40], *setype;
	Etype *etype;

	num = 0;
	for (LIST_LOOP (sub, sublist, temp)) {
		num++;

		switch (num) {
		case 1:
			etype = sub->Subscript.Expr->com.typ->etype;
			break;
		}
	}

	setype = mac_etype_short_name (etype);
	
	sprintf (var, "transpose_2_%s", setype);
	(*root)->com.str = U_str_copy (var);

	se_intrin_shape = Shape_a [2];

	se_intrin_etype = etype;

	se_intrin_typ = typ_malloc ();
	se_intrin_typ->etype = se_intrin_etype;
	se_intrin_typ->shape = se_intrin_shape;

	return 3;
}

intrin_diagonal (root)
Node **root;
{
	Node *sublist = (*root)->FunctionReference.SubscriptList;

	stand ("array intrinsic DIAGONAL");

	proc = "DIAGONAL";
	arg_num = 2;
	arg_name [0] = "ARRAY"; arg_type [0] = A_NO;
	arg_name [1] = "FILL" ; arg_type [1] = S_O;
	
	if (fill_args (sublist))
		return 0;

	rank = arg_rank [0];
	etype = arg_etype [0];

	must_be_of_rank_1 (0);

	if (arg_present [1])
		must_be_of_the_same_type (0, 1);

	setype = mac_etype_short_name (etype);
	
	sprintf (var, "diagonal_%s%d", setype, etype->kind);
	(*root)->com.str = U_str_copy (var);

	se_intrin_shape = Shape_a [2];

	se_intrin_etype = etype;

	se_intrin_typ = typ_malloc ();
	se_intrin_typ->etype = se_intrin_etype;
	se_intrin_typ->shape = se_intrin_shape;

	return 3;
}

intrin_spreads (root)
Node **root;
{
	Node *sublist = (*root)->FunctionReference.SubscriptList;

	proc = "SPREAD";
	arg_num = 3;
	arg_name [0] = "SOURCE" ; arg_type [0] = S_NO;
	arg_name [1] = "DIM"    ; arg_type [1] = S_NO;
	arg_name [2] = "NCOPIES"; arg_type [2] = S_NO;
	
	if (fill_args (sublist))
		return 0;

	rank = arg_rank [0];
	etype = arg_etype [0];

	must_be_scalar (1);
	if (must_be_integer (1))
		check_dim_no_cmf_constant (arg_node [1], 1, rank + 1);

	must_be_scalar (2);
	must_be_integer (2);
	
	setype = mac_etype_short_name (etype);
	
	sprintf (var, "spread_2_%s%d", setype, etype->kind);
	(*root)->com.str = U_str_copy (var);

	se_intrin_shape = Shape_a [1];

	se_intrin_etype = etype;

	se_intrin_typ = typ_malloc ();
	se_intrin_typ->etype = se_intrin_etype;
	se_intrin_typ->shape = se_intrin_shape;

	return 3;
}

intrin_unpack1 (root)
Node **root;
{
	Node *sublist = (*root)->FunctionReference.SubscriptList;

	proc = "UNPACK";
	arg_num = 3;
	arg_name [0] = "VECTOR"; arg_type [0] = A_NO;
	arg_name [1] = "MASK"  ; arg_type [1] = A_NO;
	arg_name [2] = "FIELD" ; arg_type [2] = S_NO;

	if (fill_args (sublist))
		return 0;
	
	etype = arg_etype [0];
	rank = arg_rank [1];

	must_be_of_rank_1 (0);
	must_be_logical (1);
	must_be_of_the_same_type (0, 2);
	must_be_conformal_to (1, 2);

	setype = mac_etype_short_name (etype);
	
	sprintf (var, "unpack_1_%d%s%d", rank, setype, etype->kind);
	(*root)->com.str = U_str_copy (var);

	decl_intrin_add (mac_etype_name (etype), (*root)->com.str);

	se_intrin_shape = Shape_a [rank];

	se_intrin_etype = etype;

	se_intrin_typ = typ_malloc ();
	se_intrin_typ->etype = se_intrin_etype;
	se_intrin_typ->shape = se_intrin_shape;

	return 3;
}

intrin_unpack2 (root)
Node **root;
{
	Node *sublist = (*root)->FunctionReference.SubscriptList;

	proc = "UNPACK";
	arg_num = 3;
	arg_name [0] = "VECTOR"; arg_type [0] = A_NO;
	arg_name [1] = "MASK"  ; arg_type [1] = A_NO;
	arg_name [2] = "FIELD" ; arg_type [2] = A_NO;
	
	if (fill_args (sublist))
		return 0;

	etype = arg_etype [0];
	rank = arg_rank [1];

	must_be_of_rank_1 (0);
	must_be_logical (1);
	must_be_of_the_same_type (0, 2);
	must_be_conformal_to (1, 2);

	setype = mac_etype_short_name (etype);
	
	sprintf (var, "unpack_2_%d%s%d", rank, setype, etype->kind);
	(*root)->com.str = U_str_copy (var);

	decl_intrin_add (mac_etype_name (etype), (*root)->com.str);

	se_intrin_shape = Shape_a [rank];

	se_intrin_etype = etype;

	se_intrin_typ = typ_malloc ();
	se_intrin_typ->etype = se_intrin_etype;
	se_intrin_typ->shape = se_intrin_shape;

	return 3;
}

static set_merge_args ()
{
	proc = "MERGE";
	arg_num = 3;
}

fill_merge_args (sublist)
Node *sublist;
{
	if (fill_args (sublist))
		return 0;

	etype = arg_etype [0];
	rank = arg_rank [0];
}

intrin_mergesss (root)
Node **root;
{
	Node *sublist = (*root)->FunctionReference.SubscriptList;

	set_merge_args ();
	arg_name [0] = "TSOURCE"; arg_type [0] = S_NO;
	arg_name [1] = "FSOURCE"; arg_type [1] = S_NO;
	arg_name [2] = "MASK"   ; arg_type [2] = S_NO;
	fill_merge_args (sublist);

	must_be_of_the_same_type (0, 1);
	must_be_logical (2);

	setype = mac_etype_short_name (etype);
	
	sprintf (var, "merge_sss_%s%d", setype, etype->kind);
	(*root)->com.str = U_str_copy (var);

	decl_intrin_add (mac_etype_name (etype), (*root)->com.str);

	se_intrin_shape = Shape_a [0];

	se_intrin_etype = etype;

	se_intrin_typ = typ_malloc ();
	se_intrin_typ->etype = se_intrin_etype;
	se_intrin_typ->shape = se_intrin_shape;

	return 3;
}

intrin_mergessa (root)  /* just to give an error */
Node **root;
{
	fatal ("argument MASK of MERGE is array for scalar TSOURCE and FSOURCE");

	return 0;
}

intrin_replicate (root)
Node **root;
{
	Node *sublist = (*root)->FunctionReference.SubscriptList;

	stand ("array intrinsic REPLICATE");

	proc = "REPLICATE";
	arg_num = 3;
	arg_name [0] = "ARRAY"  ; arg_type [0] = A_NO;
	arg_name [1] = "DIM"    ; arg_type [1] = S_NO;
	arg_name [2] = "NCOPIES"; arg_type [2] = S_NO;

	if (fill_args (sublist))
		return 0;

	rank = arg_rank [0];
	etype = arg_etype [0];

	must_not_be_c (0);

	must_be_scalar (1);
	if (must_be_integer (1) && rank)
		check_dim (arg_node [1], 1, rank);

	must_be_scalar (2);
	must_be_integer (2);

	setype = mac_etype_short_name (etype);
	
	sprintf (var, "replicate_%d%s%d", rank, setype, etype->kind);
	(*root)->com.str = U_str_copy (var);

	se_intrin_shape = Shape_a [rank];

	se_intrin_etype = etype;

	se_intrin_typ = typ_malloc ();
	se_intrin_typ->etype = se_intrin_etype;
	se_intrin_typ->shape = se_intrin_shape;

	return 3;
}

intrin_reshape_CMF (root)
Node **root;
{
	Node *sublist = (*root)->FunctionReference.SubscriptList;
	double value;

	proc = "RESHAPE";
	arg_num = 4;
	arg_name [0] = "MOLD"  ; arg_type [0] = A_NO;
	arg_name [1] = "SOURCE"; arg_type [1] = A_NO;
	arg_name [2] = "PAD"   ; arg_type [2] = A_O;
	arg_name [3] = "ORDER" ; arg_type [3] = A_O;
	
	if (fill_args (sublist))
		return 0;

	etype = arg_etype [1];

	if (arg_rank [0]  /* error recovery */
	&& !expr_is_constant (arg_node [0]->com.typ->shape->extent [0], &value)) {
		fatal ("argument MOLD of RESHAPE must be of constant size");
		rank = 1;
	}
	else {
		rank = U_nint (value);

		if (rank < 1 || rank > 7) {
			sprintf (serr
			, "argument MOLD of RESHAPE must have size in the range 1 to 7, it's size is %d"
			, rank);
			fatal (serr);
			rank = 1;
		}
	}

	must_be_of_rank_1 (0);
	must_be_integer (0);
	must_be_of_the_same_type (1, 2);
	must_be_integer (3);
	must_be_conformal_to (0, 3);

	setype = mac_etype_short_name (etype);
	
	sprintf (var, "reshapec_%d%s%d", rank, setype, etype->kind);
	(*root)->com.str = U_str_copy (var);

	se_intrin_shape = Shape_a [rank];

	se_intrin_etype = etype;

	se_intrin_typ = typ_malloc ();
	se_intrin_typ->etype = se_intrin_etype;
	se_intrin_typ->shape = se_intrin_shape;

	return 3;
}

intrin_reshape (root)
Node **root;
{
	Node *sublist = (*root)->FunctionReference.SubscriptList;
	double value;

	proc = "RESHAPE";
	arg_num = 4;
	arg_name [0] = "SOURCE"; arg_type [0] = A_NO;
	arg_name [1] = "SHAPE" ; arg_type [1] = A_NO;
	arg_name [2] = "PAD"   ; arg_type [2] = A_O;
	arg_name [3] = "ORDER" ; arg_type [3] = A_O;
	
	if (fill_args (sublist))
		return 0;

	etype = arg_etype [0];

	if (arg_rank [1]  /* error recovery */
	&&  !expr_is_constant (arg_node [1]->com.typ->shape->extent [0], &value)) {
		fatal ("argument SHAPE of RESHAPE must be of constant size");
		rank = 1;
	}
	else {
		rank = U_nint (value);

		if (rank < 1 || rank > 7) {
			sprintf (serr
			, "argument SHAPE of RESHAPE must have size in the range 1 to 7, it's size is %d"
			, rank);
			fatal (serr);
			rank = 1;
		}
	}

	must_be_of_rank_1 (1);
	must_be_integer (1);
	must_be_of_the_same_type (0, 2);
	must_be_integer (3);
	must_be_conformal_to (1, 3);

	setype = mac_etype_short_name (etype);
	
	sprintf (var, "reshape_%d%s%d", rank, setype, etype->kind);
	(*root)->com.str = U_str_copy (var);

	decl_intrin_add (mac_etype_name (etype), (*root)->com.str);

	se_intrin_shape = Shape_a [rank];

	se_intrin_etype = etype;

	se_intrin_typ = typ_malloc ();
	se_intrin_typ->etype = se_intrin_etype;
	se_intrin_typ->shape = se_intrin_shape;

	return 3;
}

intrin_packa (root)  /* this is for the one that has mask as array */
Node **root;
{
	Node *sublist = (*root)->FunctionReference.SubscriptList;
	int vector_not_present;

	proc = "PACK";
	arg_num = 3;
	arg_name [0] = "ARRAY" ; arg_type [0] = A_NO;
	arg_name [1] = "MASK"  ; arg_type [1] = A_NO;
	arg_name [2] = "VECTOR"; arg_type [2] = A_O;

	if (fill_args (sublist))
		return 0;

	rank = arg_rank [0];
	etype = arg_etype [0];

	must_be_logical (1);
	must_be_conformal_to (0, 1);

	if (arg_present [2]) {
		must_be_of_rank_1 (2);
		must_be_of_the_same_type (0, 2);
	}

	if (arg_present [2])
		vector_not_present = 0;
	else
		vector_not_present = 1;

	setype = mac_etype_short_name (etype);
	
	if (vector_not_present) {
		sprintf (var, "pack_3_%d%s%d", rank, setype, etype->kind);
		(*root)->com.str = U_str_copy (var);
	}
	else {
		sprintf (var, "pack_1_%d%s%d", rank, setype, etype->kind);
		(*root)->com.str = U_str_copy (var);
	}

	se_intrin_shape = Shape_a [1];

	se_intrin_etype = etype;

	se_intrin_typ = typ_malloc ();
	se_intrin_typ->etype = se_intrin_etype;
	se_intrin_typ->shape = se_intrin_shape;

	return 3;
}

intrin_packs (root)  /* this is for the one that has mask as scalar */
Node **root;
{
	Node *sublist = (*root)->FunctionReference.SubscriptList;
	int vector_not_present;

	proc = "PACK";
	arg_num = 3;
	arg_name [0] = "ARRAY" ; arg_type [0] = A_NO;
	arg_name [1] = "MASK"  ; arg_type [1] = S_NO;
	arg_name [2] = "VECTOR"; arg_type [2] = A_O;

	if (fill_args (sublist))
		return 0;

	rank = arg_rank [0];
	etype = arg_etype [0];

	must_be_logical (1);

	if (arg_present [2]) {
		must_be_of_rank_1 (2);
		must_be_of_the_same_type (0, 2);
	}

	if (arg_present [2])
		vector_not_present = 0;
	else
		vector_not_present = 1;

	setype = mac_etype_short_name (etype);
	
	if (vector_not_present) {
		sprintf (var, "pack_4_%d%s%d", rank, setype, etype->kind);
		(*root)->com.str = U_str_copy (var);
	}
	else {
		sprintf (var, "pack_2_%d%s%d", rank, setype, etype->kind);
		(*root)->com.str = U_str_copy (var);
	}

	se_intrin_shape = Shape_a [1];

	se_intrin_etype = etype;

	se_intrin_typ = typ_malloc ();
	se_intrin_typ->etype = se_intrin_etype;
	se_intrin_typ->shape = se_intrin_shape;

	return 3;
}

intrin_spreada (root)
Node **root;
{
	Node *sublist = (*root)->FunctionReference.SubscriptList;

	proc = "SPREAD";
	arg_num = 3;
	arg_name [0] = "SOURCE" ; arg_type [0] = A_NO;
	arg_name [1] = "DIM"    ; arg_type [1] = S_NO;
	arg_name [2] = "NCOPIES"; arg_type [2] = S_NO;
	
	if (fill_args (sublist))
		return 0;

	rank = arg_rank [0];
	etype = arg_etype [0];

	if (rank > 6) {
		sprintf (serr, "argument SOURCE of SPREAD must have rank less that 7, is %d"
		, rank);
		fatal (serr);

		rank = 1;  /* error recovery */
	}

	must_be_scalar (1);
	if (must_be_integer (1))
		check_dim_no_cmf_constant (arg_node [1], 1, rank + 1);

	must_be_scalar (2);
	must_be_integer (2);

	setype = mac_etype_short_name (etype);
	
	sprintf (var, "spread_1_%d%s%d", rank, setype, etype->kind);
	(*root)->com.str = U_str_copy (var);

	se_intrin_shape = Shape_a [rank + 1];

	se_intrin_etype = etype;

	se_intrin_typ = typ_malloc ();
	se_intrin_typ->etype = se_intrin_etype;
	se_intrin_typ->shape = se_intrin_shape;

	return 3;
}

static set_eoshift_args ()
{
	proc = "EOSHIFT";
	arg_num = 4;
	arg_name [0] = "ARRAY"   ; arg_type [0] = A_NO;

	if (flag_cmf) {
		arg_name [1] = "DIM";
		arg_type [1] = S_NO;
	}
	else {
		arg_name [3] = "DIM";
		arg_type [3] = S_O;
	}
}

static fill_eoshift_args (sublist)
Node *sublist;
{
	if (fill_args (sublist))
		return 1;

	rank = arg_rank [0];
	etype = arg_etype [0];

	if (flag_cmf) {
		must_be_integer (2);

		if (arg_present [3])
			must_be_of_the_same_type (0, 3);
	}
	else {
		must_be_integer (1);

		if (arg_present [2])
			must_be_of_the_same_type (0, 2);
	}

	return 0;
}

intrin_eoshift_ss (root)
Node **root;
{
	Node *sublist = (*root)->FunctionReference.SubscriptList;

	set_eoshift_args ();
	if (flag_cmf) {
		arg_name [2] = "SHIFT"   ; arg_type [2] = S_NO;
		arg_name [3] = "BOUNDARY"; arg_type [3] = S_O;
	}
	else {
		arg_name [1] = "SHIFT"   ; arg_type [1] = S_NO;
		arg_name [2] = "BOUNDARY"; arg_type [2] = S_O;
	}

	if (fill_eoshift_args (sublist))
		return 0;

	if (flag_cmf) {
		if (must_be_integer (1))
			if (rank)
				check_dim_no_cmf_constant (arg_node [1], 1, rank);
	}
	else {
		if (must_be_integer (3))
			if (rank && arg_present [3])
				check_dim_no_cmf_constant (arg_node [3], 1, rank);
	}

	setype = mac_etype_short_name (etype);
	
	if (flag_cmf)
		sprintf (var, "eoshiftc_ss_%d%s%d", rank, setype, etype->kind);
	else{
                if(arg_present[2])
		  sprintf (var, "eoshiftb_%d%s", rank, setype);
                else
                 sprintf (var, "eoshift_%d_%s", rank, setype);
	      }
	(*root)->com.str = U_str_copy (var);

	se_intrin_shape = Shape_a [rank];

	se_intrin_etype = etype;

	se_intrin_typ = typ_malloc ();
	se_intrin_typ->etype = se_intrin_etype;
	se_intrin_typ->shape = se_intrin_shape;

	return 3;
}

intrin_eoshift_as (root)
Node **root;
{
	Node *sublist = (*root)->FunctionReference.SubscriptList;
	int num_shift, rank_shift;

	set_eoshift_args ();
	if (flag_cmf) {
		arg_name [2] = "SHIFT"   ; arg_type [2] = A_NO;
		arg_name [3] = "BOUNDARY"; arg_type [3] = S_O;
	}
	else {
		arg_name [1] = "SHIFT"   ; arg_type [1] = A_NO;
		arg_name [2] = "BOUNDARY"; arg_type [2] = S_O;
	}

	if (fill_eoshift_args (sublist))
		return 0;

	if (flag_cmf)
		num_shift = 2;
	else
		num_shift = 1;

	rank_shift = arg_rank [num_shift];

	if (flag_cmf) {
		if (must_be_integer (1))
			if (rank)
				check_dim_no_cmf_constant (arg_node [1], 1, rank);
	}
	else {
		if (must_be_integer (3))
			if (rank && arg_present [3])
				check_dim_no_cmf_constant (arg_node [3], 1, rank);
	}

	if (rank == 1 && rank_shift)
		must_be_scalar (num_shift);

	if (rank > 1 && (rank_shift != 0 && rank_shift != (rank - 1))) {
		sprintf (serr
		, "argument SHIFT of EOSHIFT must be scalar or of rank %d, is of rank %d"
		, rank - 1, rank_shift);
		fatal (serr);
	}

	setype = mac_etype_short_name (etype);
	
	if (flag_cmf)
		sprintf (var, "eoshiftc_as_%d%s%d", rank, setype, etype->kind);
	else
		sprintf (var, "eoshift_as_%d%s%d", rank, setype, etype->kind);
	
	(*root)->com.str = U_str_copy (var);

	se_intrin_shape = Shape_a [rank];

	se_intrin_etype = etype;

	se_intrin_typ = typ_malloc ();
	se_intrin_typ->etype = se_intrin_etype;
	se_intrin_typ->shape = se_intrin_shape;

	return 3;
}

intrin_eoshift_sa (root)
Node **root;
{
	Node *sublist = (*root)->FunctionReference.SubscriptList;
	int num_boundary, rank_boundary;

	set_eoshift_args ();
	if (flag_cmf) {
		arg_name [2] = "SHIFT"   ; arg_type [2] = S_NO;
		arg_name [3] = "BOUNDARY"; arg_type [3] = A_O;
	}
	else {
		arg_name [1] = "SHIFT"   ; arg_type [1] = S_NO;
		arg_name [2] = "BOUNDARY"; arg_type [2] = A_O;
	}

	if (fill_eoshift_args (sublist))
		return 0;

	if (flag_cmf) {
		if (must_be_integer (1))
			if (rank)
				check_dim_no_cmf_constant (arg_node [1], 1, rank);
	}
	else {
		if (must_be_integer (3))
			if (rank && arg_present [3])
				check_dim_no_cmf_constant (arg_node [3], 1, rank);
	}

	if (flag_cmf)
		num_boundary = 3;
	else
		num_boundary = 2;

	rank_boundary = arg_rank [num_boundary];

	if (rank == 1 && rank_boundary) {
		sprintf (serr, "argument BOUNDARY of EOSHIFT must be scalar, is of rank %d"
		, rank_boundary);
		fatal (serr);
	}

	if (rank > 1 && (rank_boundary != 0 && rank_boundary != (rank - 1))) {
		sprintf (serr
		, "argument BOUNDARY of EOSHIFT must be scalar or of rank %d, is of rank %d"
		, rank - 1, rank_boundary);
		fatal (serr);
	}

	setype = mac_etype_short_name (etype);
	
	if (flag_cmf)
		sprintf (var, "eoshiftc_sa_%d%s%d", rank, setype, etype->kind);
	else
		sprintf (var, "eoshift_sa_%d%s%d", rank, setype, etype->kind);

	(*root)->com.str = U_str_copy (var);

	se_intrin_shape = Shape_a [rank];

	se_intrin_etype = etype;

	se_intrin_typ = typ_malloc ();
	se_intrin_typ->etype = se_intrin_etype;
	se_intrin_typ->shape = se_intrin_shape;

	return 3;
}

intrin_eoshift_aa (root)
Node **root;
{
	Node *sublist = (*root)->FunctionReference.SubscriptList;
	int rank_shift, rank_boundary, num_shift, num_boundary;

	set_eoshift_args ();
	if (flag_cmf) {
		arg_name [2] = "SHIFT"   ; arg_type [2] = A_NO;
		arg_name [3] = "BOUNDARY"; arg_type [3] = A_O;
	}
	else {
		arg_name [1] = "SHIFT"   ; arg_type [1] = A_NO;
		arg_name [2] = "BOUNDARY"; arg_type [2] = A_O;
	}

	if (fill_eoshift_args (sublist))
		return 0;

	if (flag_cmf) {
		if (must_be_integer (1))
			if (rank)
				check_dim_no_cmf_constant (arg_node [1], 1, rank);
	}
	else {
		if (must_be_integer (3))
			if (rank && arg_present [3])
				check_dim_no_cmf_constant (arg_node [3], 1, rank);
	}

	if (flag_cmf) {
		num_shift = 2;
		num_boundary = 3;
	}
	else {
		num_shift = 1;
		num_boundary = 2;
	}

	rank_shift = arg_rank [num_shift];
	rank_boundary = arg_rank [num_boundary];

	if (rank == 1 && rank_shift) {
		sprintf (serr, "argument SHIFT of EOSHIFT must be scalar, is of rank %d"
		, rank_shift);
		fatal (serr);
	}

	if (rank > 1 && (rank_shift != 0 && rank_shift != (rank - 1))) {
		sprintf (serr
		, "argument SHIFT of EOSHIFT must be scalar or of rank %d, is of rank %d"
		, rank - 1, rank_shift);
		fatal (serr);
	}

	if (rank == 1 && rank_boundary) {
		sprintf (serr, "argument BOUNDARY of EOSHIFT must be scalar, is of rank %d"
		, rank_boundary);
		fatal (serr);
	}

	if (rank > 1 && (rank_boundary != 0 && rank_boundary != (rank - 1))) {
		sprintf (serr
		, "argument BOUNDARY of EOSHIFT must be scalar or of rank %d, is of rank %d"
		, rank - 1, rank_boundary);
		fatal (serr);
	}

	setype = mac_etype_short_name (etype);

	if (flag_cmf)
		sprintf (var, "eoshiftc_aa_%d%s%d", rank, setype, etype->kind);
	else
		sprintf (var, "eoshift_aa_%d%s%d", rank, setype, etype->kind);

	(*root)->com.str = U_str_copy (var);

	se_intrin_shape = Shape_a [rank];

	se_intrin_etype = etype;

	se_intrin_typ = typ_malloc ();
	se_intrin_typ->etype = se_intrin_etype;
	se_intrin_typ->shape = se_intrin_shape;

	return 3;
}

intrin_project_fs (root)
Node **root;
{
	Node *sublist = (*root)->FunctionReference.SubscriptList;

	stand ("array intrinsic PROJECT");

	proc = "PROJECT";
	arg_num = 4;
	arg_name [0] = "ARRAY"; arg_type [0] = A_NO;
	arg_name [1] = "MASK" ; arg_type [1] = A_NO;
	arg_name [2] = "FIELD"; arg_type [2] = S_NO;
	arg_name [3] = "DIM"  ; arg_type [3] = S_O;

	if (fill_args (sublist))
		return 0;

	rank = arg_rank [0];
	etype = arg_etype [0];

	must_not_be_c (0);
	must_be_logical (1);
	must_be_of_the_same_shape (0, 1);
	must_be_of_the_same_type (0, 2);
	must_be_of_rank_x (2, rank - 1);

	if (arg_present [3])
		must_be_of_rank_x (2, rank - 1);
	else
		must_be_scalar (2);

	if (arg_present [3] && must_be_integer (3) && rank)
		check_dim (arg_node [3], 1, rank);

	setype = mac_etype_short_name (etype);
	
	if (!arg_present [3] || rank <= 1) {
		sprintf (var, "project_fs_s_%d%s%d", rank, setype, etype->kind);
		(*root)->com.str = U_str_copy (var);
		decl_intrin_add (mac_etype_name (etype), (*root)->com.str);
		se_intrin_shape = Shape_a [0];
	}
	else {
		; /* error recovery */
	}

	se_intrin_etype = etype;
	se_intrin_typ = typ_malloc ();
	se_intrin_typ->etype = se_intrin_etype;
	se_intrin_typ->shape = se_intrin_shape;

	return 3;
}

intrin_project_fa (root)
Node **root;
{
	Node *sublist = (*root)->FunctionReference.SubscriptList;

	stand ("array intrinsic PROJECT");

	proc = "PROJECT";
	arg_num = 4;
	arg_name [0] = "ARRAY"; arg_type [0] = A_NO;
	arg_name [1] = "MASK" ; arg_type [1] = A_NO;
	arg_name [2] = "FIELD"; arg_type [2] = A_NO;
	arg_name [3] = "DIM"  ; arg_type [3] = S_O;

	if (fill_args (sublist))
		return 0;

	rank = arg_rank [0];
	etype = arg_etype [0];

	must_not_be_c (0);
	must_be_logical (1);
	must_be_of_the_same_shape (0, 1);
	must_be_of_the_same_type (0, 2);

	if (arg_present [3])
		must_be_of_rank_x (2, rank - 1);
	else
		must_be_scalar (2);

	if (arg_present [3] && must_be_integer (3) && rank)
		check_dim (arg_node [3], 1, rank);

	setype = mac_etype_short_name (etype);
	
	if (!arg_present [3] || rank <= 1) {
		;  /* error recovery */
	}
	else {
		sprintf (var, "project_fa_a_%d%s%d", rank, setype, etype->kind);
		(*root)->com.str = U_str_copy (var);
		se_intrin_shape = Shape_a [rank - 1];
	}

	se_intrin_etype = etype;
	se_intrin_typ = typ_malloc ();
	se_intrin_typ->etype = se_intrin_etype;
	se_intrin_typ->shape = se_intrin_shape;

	return 3;
}

intrin_cshifta (root)
Node **root;
{
	Node *sublist = (*root)->FunctionReference.SubscriptList;
	int num_shift, rank_shift;

	proc = "CSHIFT";
	arg_num = 3;
	arg_name [0] = "ARRAY"   ; arg_type [0] = A_NO;

	if (flag_cmf) {
		arg_name [1] = "DIM"   ; arg_type [1] = S_NO;
		arg_name [2] = "SHIFT" ; arg_type [2] = A_NO;
	}
	else {
		arg_name [2] = "DIM"   ; arg_type [2] = S_O;
		arg_name [1] = "SHIFT" ; arg_type [1] = A_NO;
	}

	if (fill_args (sublist))
		return 0;

	rank = arg_rank [0];
	etype = arg_etype [0];

	if (flag_cmf)
		num_shift = 2;
	else
		num_shift = 1;

	rank_shift = arg_rank [num_shift];

	must_be_integer (num_shift);

	if (flag_cmf) {
		if (must_be_integer (1))
			if (rank)
				check_dim_no_cmf_constant (arg_node [1], 1, rank);
	}
	else {
		if (must_be_integer (2))
			if (rank & arg_present [2])
				check_dim_no_cmf_constant (arg_node [2], 1, rank);
	}

	if (rank == 1 && rank_shift) {
		sprintf (serr, "argument SHIFT of CSHIFT must be scalar, is of rank %d"
		, rank_shift);
		fatal (serr);
	}

	if (rank > 1 && (rank_shift != 0 && rank_shift != (rank - 1))) {
		sprintf (serr
		, "argument SHIFT of CSHIFT must be scalar or of rank %d, is of rank %d"
		, rank - 1, rank_shift);
		fatal (serr);
	}

	setype = mac_etype_short_name (etype);
	
	if (flag_cmf)
		sprintf (var, "cshiftc_a_%d%s%d", rank, setype, etype->kind);
	else
		sprintf (var, "cshift_a_%d%s%d", rank, setype, etype->kind);

	(*root)->com.str = U_str_copy (var);

	se_intrin_shape = Shape_a [rank];

	se_intrin_etype = etype;

	se_intrin_typ = typ_malloc ();
	se_intrin_typ->etype = se_intrin_etype;
	se_intrin_typ->shape = se_intrin_shape;

	return 3;
}

int intrin_cshifts (root)
Node **root;
{
	Node *sublist = (*root)->FunctionReference.SubscriptList;
	int num_shift, rank_shift;

	proc = "CSHIFT";
	arg_num = 3;
	arg_name [0] = "ARRAY"   ; arg_type [0] = A_NO;

	if (flag_cmf) {
		arg_name [1] = "DIM"   ; arg_type [1] = S_NO;
		arg_name [2] = "SHIFT" ; arg_type [2] = S_NO;
	}
	else {
		arg_name [2] = "DIM"   ; arg_type [2] = S_O;
		arg_name [1] = "SHIFT" ; arg_type [1] = S_NO;
	}

	if (fill_args (sublist))
		return 0;

	rank = arg_rank [0];
	etype = arg_etype [0];

	if (flag_cmf)
		num_shift = 2;
	else
		num_shift = 1;

	rank_shift = arg_rank [num_shift];

	must_be_integer (num_shift);

	if (flag_cmf) {
		if (must_be_integer (1))
			if (rank)
				check_dim_no_cmf_constant (arg_node [1], 1, rank);
	}
	else {
		if (must_be_integer (2))
			if (rank & arg_present [2])
				check_dim_no_cmf_constant (arg_node [2], 1, rank);
	}
	setype = mac_etype_short_name (etype);
	
	if (flag_cmf)
		sprintf (var, "cshiftc_s_%d%s%d", rank, setype, etype->kind);
	else
		sprintf (var, "cshift_%d_%s", rank, setype);

	(*root)->com.str = U_str_copy (var);

	se_intrin_shape = Shape_a [rank];

	se_intrin_etype = etype;

	se_intrin_typ = typ_malloc ();
	se_intrin_typ->etype = se_intrin_etype;
	se_intrin_typ->shape = se_intrin_shape;

	return 3;
}

intrin_mergeaaa (root)
Node **root;
{
	Node *sublist = (*root)->FunctionReference.SubscriptList;

	set_merge_args ();
	arg_name [0] = "TSOURCE"; arg_type [0] = A_NO;
	arg_name [1] = "FSOURCE"; arg_type [1] = A_NO;
	arg_name [2] = "MASK"   ; arg_type [2] = A_NO;
	fill_merge_args (sublist);

	must_be_of_the_same_type (0, 1);
	must_be_logical (2);

	setype = mac_etype_short_name (etype);
	
	sprintf (var, "merge_aaa_%d%s%d", rank, setype, etype->kind);
	(*root)->com.str = U_str_copy (var);

	decl_intrin_add (mac_etype_name (etype), (*root)->com.str);

	se_intrin_shape = Shape_a [rank];

	se_intrin_etype = etype;

	se_intrin_typ = typ_malloc ();
	se_intrin_typ->etype = se_intrin_etype;
	se_intrin_typ->shape = se_intrin_shape;

	return 3;
}

intrin_mergeaas (root)
Node **root;
{
	Node *sublist = (*root)->FunctionReference.SubscriptList;

	set_merge_args ();
	arg_name [0] = "TSOURCE"; arg_type [0] = A_NO;
	arg_name [1] = "FSOURCE"; arg_type [1] = A_NO;
	arg_name [2] = "MASK"   ; arg_type [2] = S_NO;
	fill_merge_args (sublist);

	must_be_of_the_same_type (0, 1);
	must_be_logical (2);

	rank = arg_rank [0];

	setype = mac_etype_short_name (etype);
	
	sprintf (var, "merge_aas_%d%s%d", rank, setype, etype->kind);
	(*root)->com.str = U_str_copy (var);

	decl_intrin_add (mac_etype_name (etype), (*root)->com.str);

	se_intrin_shape = Shape_a [rank];

	se_intrin_etype = etype;

	se_intrin_typ = typ_malloc ();
	se_intrin_typ->etype = se_intrin_etype;
	se_intrin_typ->shape = se_intrin_shape;

	return 3;
}

intrin_mergeasa (root)
Node **root;
{
	Node *sublist = (*root)->FunctionReference.SubscriptList;

	set_merge_args ();
	arg_name [0] = "TSOURCE"; arg_type [0] = A_NO;
	arg_name [1] = "FSOURCE"; arg_type [1] = S_NO;
	arg_name [2] = "MASK"   ; arg_type [2] = A_NO;
	fill_merge_args (sublist);

	must_be_of_the_same_type (0, 1);
	must_be_logical (2);

	rank = arg_rank [0];

	setype = mac_etype_short_name (etype);
	
	sprintf (var, "merge_asa_%d%s%d", rank, setype, etype->kind);
	(*root)->com.str = U_str_copy (var);

	decl_intrin_add (mac_etype_name (etype), (*root)->com.str);

	se_intrin_shape = Shape_a [rank];

	se_intrin_etype = etype;

	se_intrin_typ = typ_malloc ();
	se_intrin_typ->etype = se_intrin_etype;
	se_intrin_typ->shape = se_intrin_shape;

	return 3;
}

intrin_mergeass (root)
Node **root;
{
	Node *sublist = (*root)->FunctionReference.SubscriptList;

	set_merge_args ();
	arg_name [0] = "TSOURCE"; arg_type [0] = A_NO;
	arg_name [1] = "FSOURCE"; arg_type [1] = S_NO;
	arg_name [2] = "MASK"   ; arg_type [2] = S_NO;
	fill_merge_args (sublist);

	must_be_of_the_same_type (0, 1);
	must_be_logical (2);

	rank = arg_rank [0];

	setype = mac_etype_short_name (etype);
	
	sprintf (var, "merge_ass_%d%s%d", rank, setype, etype->kind);
	(*root)->com.str = U_str_copy (var);

	decl_intrin_add (mac_etype_name (etype), (*root)->com.str);

	se_intrin_shape = Shape_a [rank];

	se_intrin_etype = etype;

	se_intrin_typ = typ_malloc ();
	se_intrin_typ->etype = se_intrin_etype;
	se_intrin_typ->shape = se_intrin_shape;

	return 3;
}

intrin_mergesaa (root)
Node **root;
{
	Node *sublist = (*root)->FunctionReference.SubscriptList;

	set_merge_args ();
	arg_name [0] = "TSOURCE"; arg_type [0] = S_NO;
	arg_name [1] = "FSOURCE"; arg_type [1] = A_NO;
	arg_name [2] = "MASK"   ; arg_type [2] = A_NO;
	fill_merge_args (sublist);

	must_be_of_the_same_type (0, 1);
	must_be_logical (2);

	rank = arg_rank [1];

	setype = mac_etype_short_name (etype);
	
	sprintf (var, "merge_saa_%d%s%d", rank, setype, etype->kind);
	(*root)->com.str = U_str_copy (var);

	decl_intrin_add (mac_etype_name (etype), (*root)->com.str);

	se_intrin_shape = Shape_a [rank];

	se_intrin_etype = etype;

	se_intrin_typ = typ_malloc ();
	se_intrin_typ->etype = se_intrin_etype;
	se_intrin_typ->shape = se_intrin_shape;

	return 3;
}

intrin_mergesas (root)
Node **root;
{
	Node *sublist = (*root)->FunctionReference.SubscriptList;

	set_merge_args ();
	arg_name [0] = "TSOURCE"; arg_type [0] = S_NO;
	arg_name [1] = "FSOURCE"; arg_type [1] = A_NO;
	arg_name [2] = "MASK"   ; arg_type [2] = S_NO;
	fill_merge_args (sublist);

	must_be_of_the_same_type (0, 1);
	must_be_logical (2);

	rank = arg_rank [1];

	setype = mac_etype_short_name (etype);
	
	sprintf (var, "merge_sas_%d%s%d", rank, setype, etype->kind);
	(*root)->com.str = U_str_copy (var);

	decl_intrin_add (mac_etype_name (etype), (*root)->com.str);

	se_intrin_shape = Shape_a [rank];

	se_intrin_etype = etype;

	se_intrin_typ = typ_malloc ();
	se_intrin_typ->etype = se_intrin_etype;
	se_intrin_typ->shape = se_intrin_shape;

	return 3;
}

intrin_maxloc_a (root)
Node **root;
{
	Node *sublist = (*root)->FunctionReference.SubscriptList;

	proc = "MAXLOC";
	arg_num = 2;
	arg_name [0] = "ARRAY"; arg_type [0] = A_NO;
	arg_name [1] = "MASK";  arg_type [1] = A_O;

	if (fill_args (sublist))
		return 0;

	rank = arg_rank [0];
	etype = arg_etype [0];

	setype = mac_etype_short_name (etype);

	if (arg_present [1]) {
		must_be_logical (1);
		must_be_conformal_to (0, 1);
	}

	must_not_be_lcx (0);
        if(arg_present[1])
          sprintf (var, "maxloc_%d_%s_M", rank, setype);
        else
          sprintf (var, "maxloc_%d_%s", rank, setype);
	(*root)->com.str = U_str_copy (var);

	se_intrin_shape = Shape_a [1];

	se_intrin_etype = Etype_INTEGER_0;

	se_intrin_typ = typ_malloc ();
	se_intrin_typ->etype = se_intrin_etype;
	se_intrin_typ->shape = se_intrin_shape;

	return 3;
}

intrin_maxloc_s (root)
Node **root;
{
	Node *sublist = (*root)->FunctionReference.SubscriptList;

	proc = "MAXLOC";
	arg_num = 2;
	arg_name [0] = "ARRAY"; arg_type [0] = A_NO;
	arg_name [1] = "MASK";  arg_type [1] = S_O;

	if (fill_args (sublist))
		return 0;

	rank = arg_rank [0];
	etype = arg_etype [0];

	setype = mac_etype_short_name (etype);

	if (arg_present [1])
		must_be_logical (1);

	must_not_be_lcx (0);

	sprintf (var, "maxloc_s_%d%s%d", rank, setype, etype->kind);
	(*root)->com.str = U_str_copy (var);

	se_intrin_shape = Shape_a [1];

	se_intrin_etype = Etype_INTEGER_0;

	se_intrin_typ = typ_malloc ();
	se_intrin_typ->etype = se_intrin_etype;
	se_intrin_typ->shape = se_intrin_shape;

	return 3;
}

intrin_minloc_a (root)
Node **root;
{
	Node *sublist = (*root)->FunctionReference.SubscriptList;

	proc = "MINLOC";
	arg_num = 2;
	arg_name [0] = "ARRAY"; arg_type [0] = A_NO;
	arg_name [1] = "MASK";  arg_type [1] = A_O;

	if (fill_args (sublist))
		return 0;

	rank = arg_rank [0];
	etype = arg_etype [0];

	setype = mac_etype_short_name (etype);

	if (arg_present [1]) {
		must_be_logical (1);
		must_be_conformal_to (0, 1);
	}

	must_not_be_lcx (0);

        if(arg_present[1])
             sprintf (var, "minloc_%d_%s_M", rank, setype);
        else
             sprintf (var, "minloc_%d_%s", rank, setype);
	(*root)->com.str = U_str_copy (var);

	se_intrin_shape = Shape_a [1];

	se_intrin_etype = Etype_INTEGER_0;

	se_intrin_typ = typ_malloc ();
	se_intrin_typ->etype = se_intrin_etype;
	se_intrin_typ->shape = se_intrin_shape;

	return 3;
}

intrin_minloc_s (root)
Node **root;
{
	Node *sublist = (*root)->FunctionReference.SubscriptList;

	proc = "MINLOC";
	arg_num = 2;
	arg_name [0] = "ARRAY"; arg_type [0] = A_NO;
	arg_name [1] = "MASK";  arg_type [1] = S_O;

	if (fill_args (sublist))
		return 0;

	rank = arg_rank [0];
	etype = arg_etype [0];

	setype = mac_etype_short_name (etype);

	if (arg_present [1])
		must_be_logical (1);

	must_not_be_lcx (0);

        if(arg_present[1])
             sprintf (var, "minloc_%d_%s_M", rank, setype);
        else
             sprintf (var, "minloc_%d_%s", rank, setype);

	(*root)->com.str = U_str_copy (var);

	se_intrin_shape = Shape_a [1];

	se_intrin_etype = Etype_INTEGER_0;

	se_intrin_typ = typ_malloc ();
	se_intrin_typ->etype = se_intrin_etype;
	se_intrin_typ->shape = se_intrin_shape;

	return 3;
}

intrin_maxval_a (root)
Node **root;
{
	Node *sublist = (*root)->FunctionReference.SubscriptList;

	proc = "MAXVAL";
	arg_num = 3;
	arg_name [0] = "ARRAY"; arg_type [0] = A_NO;
	arg_name [1] = "DIM"  ; arg_type [1] = S_O;
	arg_name [2] = "MASK" ; arg_type [2] = A_O;

	if (fill_args (sublist))
		return 0;

	rank = arg_rank [0];
	etype = arg_etype [0];

	setype = mac_etype_short_name (etype);

	must_not_be_lcx (0);

	if (arg_present [1] && must_be_integer (1))
		if (rank)
			check_dim_no_cmf_constant (arg_node [1], 1, rank);

	if (arg_present [2]) {
		must_be_logical (2);
		must_be_conformal_to (0, 2);
	}

	if (arg_present [1])
		dim_not_present = 0;
	else
		dim_not_present = 1;

	setype = mac_etype_short_name (etype);
	
	if (dim_not_present || rank == 1) {
                if(arg_present[2])
		  sprintf (var, "maxval_%d_%s_M", rank, setype);
                else		
                    sprintf (var, "maxval_%d_%s", rank, setype);

		(*root)->com.str = U_str_copy (var);

		decl_intrin_add (mac_etype_name (etype), (*root)->com.str);

		se_intrin_shape = Shape_a [0];
	}
	else {
                if(dim_not_present){
                      if(arg_present[2])
	        	  sprintf (var, "maxval_%d_%s_s_M", rank, setype);
                      else		
                          sprintf (var, "maxval_%d_%s_s", rank, setype);
		    }
                else {
                      if(arg_present[2])
	        	  sprintf (var, "maxval_%d_%s_v_M", rank, setype);
                      else		
                          sprintf (var, "maxval_%d_%s_v", rank, setype);
	        }
      

		(*root)->com.str = U_str_copy (var);

		se_intrin_shape = Shape_a [rank - 1];
      }

	se_intrin_etype = etype;

	se_intrin_typ = typ_malloc ();
	se_intrin_typ->etype = se_intrin_etype;
	se_intrin_typ->shape = se_intrin_shape;

	return 3;
}

intrin_maxval_s (root)
Node **root;
{
	Node *sublist = (*root)->FunctionReference.SubscriptList;

	proc = "MAXVAL";
	arg_num = 3;
	arg_name [0] = "ARRAY"; arg_type [0] = A_NO;
	arg_name [1] = "DIM"  ; arg_type [1] = S_O;
	arg_name [2] = "MASK" ; arg_type [2] = S_O;

	if (fill_args (sublist))
		return 0;

	rank = arg_rank [0];
	etype = arg_etype [0];

	setype = mac_etype_short_name (etype);

	must_not_be_lcx (0);

	if (arg_present [1] && must_be_integer (1))
		if (rank)
			check_dim_no_cmf_constant (arg_node [1], 1, rank);

	if (arg_present [2])
		must_be_logical (2);

	if (arg_present [1])
		dim_not_present = 0;
	else
		dim_not_present = 1;

	setype = mac_etype_short_name (etype);
	
        if (dim_not_present || rank == 1) {
                if(arg_present[2])
		  sprintf (var, "maxval_%d_%s_M", rank, setype);
                else		
                    sprintf (var, "maxval_%d_%s", rank, setype);

		(*root)->com.str = U_str_copy (var);

		decl_intrin_add (mac_etype_name (etype), (*root)->com.str);

		se_intrin_shape = Shape_a [0];
	}
	else {
                if(dim_not_present){
                      if(arg_present[2])
	        	  sprintf (var, "maxval_%d_%s_s_M", rank, setype);
                      else		
                          sprintf (var, "maxval_%d_%s_s", rank, setype);
		    }
                else {
                      if(arg_present[2])
	        	  sprintf (var, "maxval_%d_%s_v_M", rank, setype);
                      else		
                          sprintf (var, "maxval_%d_%s_v", rank, setype);
 	        }

		(*root)->com.str = U_str_copy (var);

		se_intrin_shape = Shape_a [rank - 1];
	}

	se_intrin_etype = etype;

	se_intrin_typ = typ_malloc ();
	se_intrin_typ->etype = se_intrin_etype;
	se_intrin_typ->shape = se_intrin_shape;

	return 3;
}

intrin_minval_a (root)
Node **root;
{
	Node *sublist = (*root)->FunctionReference.SubscriptList;

	proc = "MINVAL";
	arg_num = 3;
	arg_name [0] = "ARRAY"; arg_type [0] = A_NO;
	arg_name [1] = "DIM";   arg_type [1] = S_O;
	arg_name [2] = "MASK";  arg_type [2] = A_O;

	if (fill_args (sublist))
		return 0;

	rank = arg_rank [0];
	etype = arg_etype [0];

	setype = mac_etype_short_name (etype);

	must_not_be_lcx (0);

	if (arg_present [1] && must_be_integer (1))
		if (rank)
			check_dim_no_cmf_constant (arg_node [1], 1, rank);

	if (arg_present [2]) {
		must_be_logical (2);
		must_be_conformal_to (0, 2);
	}

	if (arg_present [1])
		dim_not_present = 0;
	else
		dim_not_present = 1;

	setype = mac_etype_short_name (etype);
	
        if (dim_not_present || rank == 1) {
                if(arg_present[2])
		  sprintf (var, "minval_%d_%s_M", rank, setype);
                else		
                    sprintf (var, "minval_%d_%s", rank, setype);

		(*root)->com.str = U_str_copy (var);

		decl_intrin_add (mac_etype_name (etype), (*root)->com.str);

		se_intrin_shape = Shape_a [0];
	}
	else {
                if(dim_not_present){
                      if(arg_present[2])
	        	  sprintf (var, "minval_%d_%s_s_M", rank, setype);
                      else		
                          sprintf (var, "minval_%d_%s_s", rank, setype);
		    }
                else {
                      if(arg_present[2])
	        	  sprintf (var, "minval_%d_%s_v_M", rank, setype);
                      else		
                          sprintf (var, "minval_%d_%s_v", rank, setype);
 	        }

		(*root)->com.str = U_str_copy (var);

		se_intrin_shape = Shape_a [rank - 1];
	}

	se_intrin_etype = etype;

	se_intrin_typ = typ_malloc ();
	se_intrin_typ->etype = se_intrin_etype;
	se_intrin_typ->shape = se_intrin_shape;

	return 3;
}

intrin_minval_s (root)
Node **root;
{
	Node *sublist = (*root)->FunctionReference.SubscriptList;

	proc = "MINVAL";
	arg_num = 3;
	arg_name [0] = "ARRAY"; arg_type [0] = A_NO;
	arg_name [1] = "DIM";   arg_type [1] = S_O;
	arg_name [2] = "MASK";  arg_type [2] = S_O;

	if (fill_args (sublist))
		return 0;

	rank = arg_rank [0];
	etype = arg_etype [0];

	setype = mac_etype_short_name (etype);

	must_not_be_lcx (0);

	if (arg_present [1] && must_be_integer (1))
		if (rank)
			check_dim_no_cmf_constant (arg_node [1], 1, rank);

	if (arg_present [2])
		must_be_logical (2);

	if (arg_present [1])
		dim_not_present = 0;
	else
		dim_not_present = 1;

	setype = mac_etype_short_name (etype);
	
        if (dim_not_present || rank == 1) {
                if(arg_present[2])
		  sprintf (var, "minval_%d_%s_M", rank, setype);
                else		
                    sprintf (var, "minval_%d_%s", rank, setype);

		(*root)->com.str = U_str_copy (var);

		decl_intrin_add (mac_etype_name (etype), (*root)->com.str);

		se_intrin_shape = Shape_a [0];
	}
	else {
                if(dim_not_present){
                      if(arg_present[2])
	        	  sprintf (var, "minval_%d_%s_s_M", rank, setype);
                      else		
                          sprintf (var, "minval_%d_%s_s", rank, setype);
		    }
                else {
                      if(arg_present[2])
	        	  sprintf (var, "minval_%d_%s_v_M", rank, setype);
                      else		
                          sprintf (var, "minval_%d_%s_v", rank, setype);
 	        }

		(*root)->com.str = U_str_copy (var);

		se_intrin_shape = Shape_a [rank - 1];
	}


	se_intrin_etype = etype;

	se_intrin_typ = typ_malloc ();
	se_intrin_typ->etype = se_intrin_etype;
	se_intrin_typ->shape = se_intrin_shape;

	return 3;
}

intrin_producta (root)
Node **root;
{
	Node *sublist = (*root)->FunctionReference.SubscriptList;

	proc = "PRODUCT";
	arg_num = 3;
	arg_name [0] = "ARRAY"; arg_type [0] = A_NO;
	arg_name [1] = "DIM"  ; arg_type [1] = S_O;
	arg_name [2] = "MASK" ; arg_type [2] = A_NO;
	
	if (fill_args (sublist))
		return 0;

	rank = arg_rank [0];
	etype = arg_etype [0];

	must_not_be_lc (0);

	if (arg_present [1] && must_be_integer (1) && rank)
		check_dim (arg_node [1], 1, rank);

	if (arg_present [2]) {
		must_be_logical (2);
		must_be_conformal_to (0, 2);
	}

	if (arg_present [1])
		dim_not_present = 0;
	else
		dim_not_present = 1;

	setype = mac_etype_short_name (etype);
	
	if (dim_not_present || rank == 1) {
                if(arg_present[2])
		  sprintf (var, "product_%d_%s_M", rank, setype);
                else		
                    sprintf (var, "product_%d_%s", rank, setype);

		(*root)->com.str = U_str_copy (var);

		decl_intrin_add (mac_etype_name (etype), (*root)->com.str);

		se_intrin_shape = Shape_a [0];
	}
	else {
                if(dim_not_present){
                      if(arg_present[2])
	        	  sprintf (var, "product_%d_%s_s_M", rank, setype);
                      else		
                          sprintf (var, "product_%d_%s_s", rank, setype);
		    }
                else {
                      if(arg_present[2])
	        	  sprintf (var, "product_%d_%s_v_M", rank, setype);
                      else		
                          sprintf (var, "product_%d_%s_v", rank, setype);
	        }
      

		(*root)->com.str = U_str_copy (var);

		se_intrin_shape = Shape_a [rank - 1];
      }


	se_intrin_etype = etype;

	se_intrin_typ = typ_malloc ();
	se_intrin_typ->etype = se_intrin_etype;
	se_intrin_typ->shape = se_intrin_shape;

	return 3;
}

intrin_products (root)
Node **root;
{
	Node *sublist = (*root)->FunctionReference.SubscriptList;

	proc = "PRODUCT";
	arg_num = 3;
	arg_name [0] = "ARRAY"; arg_type [0] = A_NO;
	arg_name [1] = "DIM"  ; arg_type [1] = S_O;
	arg_name [2] = "MASK" ; arg_type [2] = S_NO;
	
	if (fill_args (sublist))
		return 0;

	rank = arg_rank [0];
	etype = arg_etype [0];

	must_not_be_lc (0);

	if (arg_present [1] && must_be_integer (1) && rank)
		check_dim (arg_node [1], 1, rank);

	if (arg_present [2])
		must_be_logical (2);

	if (arg_present [1])
		dim_not_present = 0;
	else
		dim_not_present = 1;

	setype = mac_etype_short_name (etype);
	
	if (dim_not_present || rank == 1) {
                if(arg_present[2])
		  sprintf (var, "product_%d_%s_M", rank, setype);
                else		
                    sprintf (var, "product_%d_%s", rank, setype);

		(*root)->com.str = U_str_copy (var);

		decl_intrin_add (mac_etype_name (etype), (*root)->com.str);

		se_intrin_shape = Shape_a [0];
	}
	else {
                if(dim_not_present){
                      if(arg_present[2])
	        	  sprintf (var, "product_%d_%s_s_M", rank, setype);
                      else		
                          sprintf (var, "product_%d_%s_s", rank, setype);
		    }
                else {
                      if(arg_present[2])
	        	  sprintf (var, "product_%d_%s_v_M", rank, setype);
                      else		
                          sprintf (var, "product_%d_%s_v", rank, setype);
	        }
      

		(*root)->com.str = U_str_copy (var);

		se_intrin_shape = Shape_a [rank - 1];
        }

	se_intrin_etype = etype;

	se_intrin_typ = typ_malloc ();
	se_intrin_typ->etype = se_intrin_etype;
	se_intrin_typ->shape = se_intrin_shape;

	return 3;
}

intrin_suma (root)
Node **root;
{
	Node *sublist = (*root)->FunctionReference.SubscriptList;

	proc = "SUM";
	arg_num = 3;
	arg_name [0] = "ARRAY"; arg_type [0] = A_NO;
	arg_name [1] = "DIM"  ; arg_type [1] = S_O;
	arg_name [2] = "MASK" ; arg_type [2] = A_NO;
	
	if (fill_args (sublist))
		return 0;

	rank = arg_rank [0];
	etype = arg_etype [0];

	must_not_be_lc (0);

	if (arg_present [1] && must_be_integer (1) && rank)
		check_dim (arg_node [1], 1, rank);

	if (arg_present [2]) {
		must_be_logical (2);
		must_be_conformal_to (0, 2);
	}

	if (arg_present [1])
		dim_not_present = 0;
	else
		dim_not_present = 1;

	setype = mac_etype_short_name (etype);
	
	if (dim_not_present || rank == 1) {
                if(arg_present[2])
		  sprintf (var, "sum_%d_%s_s_M", rank, setype);
                else		
                    sprintf (var, "sum_%d_%s_s", rank, setype);

		(*root)->com.str = U_str_copy (var);

		decl_intrin_add (mac_etype_name (etype), (*root)->com.str);

		se_intrin_shape = Shape_a [0];
	}
	else {
                if(dim_not_present){
                      if(arg_present[2])
	        	  sprintf (var, "sum_%d_%s_s_M", rank, setype);
                      else		
                          sprintf (var, "sum_%d_%s_s", rank, setype);
		    }
                else {
                      if(arg_present[2])
	        	  sprintf (var, "sum_%d_%s_v_M", rank, setype);
                      else		
                          sprintf (var, "sum_%d_%s_v", rank, setype);
	        }
      

		(*root)->com.str = U_str_copy (var);

		se_intrin_shape = Shape_a [rank - 1];
        }

	se_intrin_etype = etype;

	se_intrin_typ = typ_malloc ();
	se_intrin_typ->etype = se_intrin_etype;
	se_intrin_typ->shape = se_intrin_shape;

	return 3;
}

intrin_sums (root)
Node **root;
{
	Node *sublist = (*root)->FunctionReference.SubscriptList;

	proc = "SUM";
	arg_num = 3;
	arg_name [0] = "ARRAY"; arg_type [0] = A_NO;
	arg_name [1] = "DIM"  ; arg_type [1] = S_O;
	arg_name [2] = "MASK" ; arg_type [2] = S_NO;
	
	if (fill_args (sublist))
		return 0;

	rank = arg_rank [0];
	etype = arg_etype [0];

	must_not_be_lc (0);

	if (arg_present [1] && must_be_integer (1) && rank)
		check_dim (arg_node [1], 1, rank);

	if (arg_present [2])
		must_be_logical (2);

	if (arg_present [1])
		dim_not_present = 0;
	else
		dim_not_present = 1;

	setype = mac_etype_short_name (etype);
	
	if (dim_not_present || rank == 1) {
                if(arg_present[2])
		  sprintf (var, "sum_%d_%s_s_M", rank, setype);
                else		
                    sprintf (var, "sum_%d_%s_s", rank, setype);

		(*root)->com.str = U_str_copy (var);

		decl_intrin_add (mac_etype_name (etype), (*root)->com.str);

		se_intrin_shape = Shape_a [0];
	}
	else {
                if(dim_not_present){
                      if(arg_present[2])
	        	  sprintf (var, "sum_%d_%s_s_M", rank, setype);
                      else		
                          sprintf (var, "sum_%d_%s_s", rank, setype);
		    }
                else {
                      if(arg_present[2])
	        	  sprintf (var, "sum_%d_%s_v_M", rank, setype);
                      else		
                          sprintf (var, "sum_%d_%s_v", rank, setype);
	        }
      

		(*root)->com.str = U_str_copy (var);

		se_intrin_shape = Shape_a [rank - 1];
        }

	se_intrin_etype = etype;

	se_intrin_typ = typ_malloc ();
	se_intrin_typ->etype = se_intrin_etype;
	se_intrin_typ->shape = se_intrin_shape;

	return 3;
}

intrin_set_intent ()
{
	T_Entry *entry;

	for_TABLE_LOOP_glo (entry)
		if (IS_intrinsic (entry) && !IS_f77_intrinsic (entry))
			interf_set_intent_in (entry);
}

intrin_init ()  /* puts all intrinsics in the symbol table */
{
	T_Entry *entry;

	if (flag_no_intrin)
		return;

	entry = install_glo ("present", TAB_NAME, 0);
	SET_compile_time (entry);
	SET_intrinsic (entry);
	SET_no_checking (entry);
	entry->compile_time = intrin_present;

	entry = install_glo ("@rank", TAB_NAME, 0);
	SET_compile_time (entry);
	SET_intrinsic (entry);
	SET_no_checking (entry);
	entry->compile_time = intrin_a_rank;

	entry = install_glo ("rank", TAB_NAME, 0);
	SET_compile_time (entry);
	SET_intrinsic (entry);
	SET_no_checking (entry);
	entry->compile_time = intrin_rank;

	entry = install_glo ("@size", TAB_NAME, 0);
	SET_compile_time (entry);
	SET_intrinsic (entry);
	SET_no_checking (entry);
	entry->compile_time = intrin_size;

	entry = install_glo ("size", TAB_NAME, 0);
	SET_intrinsic (entry);
	SET_no_checking (entry);
	SET_compile_time (entry);
	entry->compile_time = intrin_size_F90;
	entry = install_glo ("dsize", TAB_NAME, 0);
	SET_intrinsic (entry);
	SET_no_checking (entry);
	SET_compile_time (entry);
	entry->compile_time = intrin_size_CMF;

	entry = install_glo ("shape", TAB_NAME, 0);
	SET_intrinsic (entry);
	SET_no_checking (entry);
	entry = install_glo ("dshape", TAB_NAME, 0);
	SET_intrinsic (entry);
	SET_no_checking (entry);
	entry = install_glo ("shape_s", TAB_NAME, 0);
	SET_intrinsic (entry);
	SET_no_checking (entry);
	SET_compile_time (entry);
	entry->compile_time = intrin_shape_s_F90;
	entry = install_glo ("shape_a", TAB_NAME, 0);
	SET_intrinsic (entry);
	SET_no_checking (entry);
	SET_compile_time (entry);
	entry->compile_time = intrin_shape_a_F90;
	entry = install_glo ("dshape_s", TAB_NAME, 0);
	SET_intrinsic (entry);
	SET_no_checking (entry);
	SET_compile_time (entry);
	entry->compile_time = intrin_shape_s_CMF;
	entry = install_glo ("dshape_a", TAB_NAME, 0);
	SET_intrinsic (entry);
	SET_no_checking (entry);
	SET_compile_time (entry);
	entry->compile_time = intrin_shape_a_CMF;

	entry = install_glo ("@lbound", TAB_NAME, 0);
	SET_compile_time (entry);
	SET_intrinsic (entry);
	SET_no_checking (entry);
	entry->compile_time = intrin_a_lbound;

	entry = install_glo ("@ubound", TAB_NAME, 0);
	SET_compile_time (entry);
	SET_intrinsic (entry);
	SET_no_checking (entry);
	entry->compile_time = intrin_a_ubound;

	entry = install_glo ("iand", TAB_NAME, 0);
	SET_compile_time (entry);
	SET_intrinsic (entry);
	SET_elemental (entry);
	SET_f77_intrinsic (entry);
	entry->compile_time = intrin_iand;
	entry = install_glo ("ieor", TAB_NAME, 0);
	SET_compile_time (entry);
	SET_intrinsic (entry);
	SET_elemental (entry);
	SET_f77_intrinsic (entry);
	entry->compile_time = intrin_ieor;
	entry = install_glo ("ior", TAB_NAME, 0);
	SET_compile_time (entry);
	SET_intrinsic (entry);
	SET_elemental (entry);
	SET_f77_intrinsic (entry);
	entry->compile_time = intrin_ior;

	entry = install_glo ("dcmplx", TAB_NAME, 0);
	SET_compile_time (entry);
	SET_intrinsic (entry);
	SET_no_checking (entry);
	SET_elemental (entry);
	SET_f77_intrinsic (entry);
	entry->compile_time = intrin_cmplx;

	entry = install_glo ("cmplx", TAB_NAME, 0);
	SET_compile_time (entry);
	SET_intrinsic (entry);
	SET_no_checking (entry);
	SET_elemental (entry);
	SET_f77_intrinsic (entry);
	entry->compile_time = intrin_cmplx;

	entry = install_glo ("aimag", TAB_NAME, 0);
	SET_compile_time (entry);
	SET_intrinsic (entry);
	SET_no_checking (entry);
	SET_elemental (entry);
	SET_f77_intrinsic (entry);
	entry->compile_time = intrin_aimag;

	entry = install_glo ("conjg", TAB_NAME, 0);
	SET_compile_time (entry);
	SET_intrinsic (entry);
	SET_no_checking (entry);
	SET_elemental (entry);
	SET_f77_intrinsic (entry);
	entry->compile_time = intrin_conjg;

	entry = install_glo ("dimag", TAB_NAME, 0);
	SET_intrinsic (entry);
	SET_elemental (entry);
	SET_f77_intrinsic (entry);

	entry = install_glo ("dconjg", TAB_NAME, 0);
	SET_intrinsic (entry);
	SET_elemental (entry);
	SET_f77_intrinsic (entry);

	entry = install_glo ("rand", TAB_NAME, 0);
	SET_elemental (entry);
	SET_intrinsic (entry);
	SET_f77_intrinsic (entry);

	entry = install_glo ("ielem", TAB_NAME, 0);  /*!!! for testing */
	SET_elemental (entry);
	SET_intrinsic (entry);
	SET_no_checking (entry);

/* array intrinsics preprocessed by compile time */

	entry = install_glo ("firstloc", TAB_NAME, 0);
	SET_compile_time (entry);
	entry->compile_time = intrin_firstloc;
	SET_intrinsic (entry);
	entry = install_glo ("lastloc", TAB_NAME, 0);
	SET_compile_time (entry);
	entry->compile_time = intrin_lastloc;
	SET_intrinsic (entry);

	entry = install_glo ("all", TAB_NAME, 0);
	SET_compile_time (entry);
	entry->compile_time = intrin_all;
	SET_intrinsic (entry);

        entry = install_glo ("any", TAB_NAME, 0);
	SET_compile_time (entry);
	entry->compile_time = intrin_any;
	SET_intrinsic (entry);

	entry = install_glo ("count", TAB_NAME, 0);
	SET_compile_time (entry);
	entry->compile_time = intrin_count;
	SET_intrinsic (entry);

	entry = install_glo ("sum", TAB_NAME, 0);
	SET_intrinsic (entry);
	SET_no_checking (entry);
	entry = install_glo ("sums", TAB_NAME, 0);
	SET_compile_time (entry);
	SET_no_checking (entry);
	entry->compile_time = intrin_sums;
	SET_intrinsic (entry);
	entry = install_glo ("suma", TAB_NAME, 0);
	SET_compile_time (entry);
	SET_no_checking (entry);
	entry->compile_time = intrin_suma;
	SET_intrinsic (entry);

	entry = install_glo ("minloc", TAB_NAME, 0);
	SET_intrinsic (entry);
	SET_no_checking (entry);
	entry = install_glo ("minloc_a", TAB_NAME, 0);
	SET_compile_time (entry);
	SET_no_checking (entry);
	entry->compile_time = intrin_minloc_a;
	entry = install_glo ("minloc_s", TAB_NAME, 0);
	SET_compile_time (entry);
	SET_no_checking (entry);
	entry->compile_time = intrin_minloc_s;
	SET_intrinsic (entry);

	entry = install_glo ("maxval", TAB_NAME, 0);
	SET_intrinsic (entry);
	SET_no_checking (entry);
	entry = install_glo ("maxval_a", TAB_NAME, 0);
	SET_compile_time (entry);
	SET_no_checking (entry);
	entry->compile_time = intrin_maxval_a;
	SET_intrinsic (entry);
	entry = install_glo ("maxval_s", TAB_NAME, 0);
	SET_compile_time (entry);
	SET_no_checking (entry);
	entry->compile_time = intrin_maxval_s;
	SET_intrinsic (entry);

	entry = install_glo ("minval", TAB_NAME, 0);
	SET_intrinsic (entry);
	SET_no_checking (entry);
	entry = install_glo ("minval_a", TAB_NAME, 0);
	SET_compile_time (entry);
	SET_no_checking (entry);
	entry->compile_time = intrin_minval_a;
	SET_intrinsic (entry);
	entry = install_glo ("minval_s", TAB_NAME, 0);
	SET_compile_time (entry);
	SET_no_checking (entry);
	entry->compile_time = intrin_minval_s;
	SET_intrinsic (entry);

	entry = install_glo ("maxloc", TAB_NAME, 0);
	SET_intrinsic (entry);
	SET_no_checking (entry);
	entry = install_glo ("maxloc_a", TAB_NAME, 0);
	SET_compile_time (entry);
	SET_no_checking (entry);
	entry->compile_time = intrin_maxloc_a;
	SET_intrinsic (entry);
	entry = install_glo ("maxloc_s", TAB_NAME, 0);
	SET_compile_time (entry);
	SET_no_checking (entry);
	entry->compile_time = intrin_maxloc_s;
	SET_intrinsic (entry);

	entry = install_glo ("pack", TAB_NAME, 0);
	SET_intrinsic (entry);
	SET_no_checking (entry);
	entry = install_glo ("packa", TAB_NAME, 0);
	SET_compile_time (entry);
	SET_no_checking (entry);
	entry->compile_time = intrin_packa;
	SET_intrinsic (entry);
	entry = install_glo ("packs", TAB_NAME, 0);
	SET_compile_time (entry);
	SET_no_checking (entry);
	entry->compile_time = intrin_packs;
	SET_intrinsic (entry);

	entry = install_glo ("matmul", TAB_NAME, 0);
	SET_intrinsic (entry);
	SET_no_checking (entry);
	SET_compile_time (entry);
	entry->compile_time = intrin_matmul;

	entry = install_glo ("dot_product", TAB_NAME, 0);
	SET_intrinsic (entry);
	SET_no_checking (entry);
	SET_compile_time (entry);
	entry->compile_time = intrin_dot_product_F90;

	entry = install_glo ("dotproduct", TAB_NAME, 0);
	SET_intrinsic (entry);
	SET_no_checking (entry);
	SET_compile_time (entry);
	entry->compile_time = intrin_dotproduct_CMF;

	entry = install_glo ("transpose", TAB_NAME, 0);
	SET_intrinsic (entry);
	SET_no_checking (entry);
	SET_compile_time (entry);
	entry->compile_time = intrin_transpose;

	entry = install_glo ("diagonal", TAB_NAME, 0);
	SET_intrinsic (entry);
	SET_no_checking (entry);
	SET_compile_time (entry);
	entry->compile_time = intrin_diagonal;

	entry = install_glo ("spread", TAB_NAME, 0);
	SET_intrinsic (entry);
	SET_no_checking (entry);
	entry = install_glo ("spreada", TAB_NAME, 0);
	SET_compile_time (entry);
	SET_no_checking (entry);
	entry->compile_time = intrin_spreada;
	SET_intrinsic (entry);
	entry = install_glo ("spreads", TAB_NAME, 0);
	SET_compile_time (entry);
	SET_no_checking (entry);
	entry->compile_time = intrin_spreads;
	SET_intrinsic (entry);

	entry = install_glo ("unpack", TAB_NAME, 0);
	SET_intrinsic (entry);
	SET_no_checking (entry);
	entry = install_glo ("unpack1", TAB_NAME, 0);
	SET_compile_time (entry);
	SET_no_checking (entry);
	entry->compile_time = intrin_unpack1;
	SET_intrinsic (entry);
	entry = install_glo ("unpack2", TAB_NAME, 0);
	SET_compile_time (entry);
	SET_no_checking (entry);
	entry->compile_time = intrin_unpack2;
	SET_intrinsic (entry);

	entry = install_glo ("eoshift", TAB_NAME, 0);
	SET_intrinsic (entry);
	SET_no_checking (entry);
	entry = install_glo ("eoshift_ss", TAB_NAME, 0);
	SET_compile_time (entry);
	SET_no_checking (entry);
	entry->compile_time = intrin_eoshift_ss;
	SET_intrinsic (entry);
	entry = install_glo ("eoshift_as", TAB_NAME, 0);
	SET_compile_time (entry);
	SET_no_checking (entry);
	entry->compile_time = intrin_eoshift_as;
	SET_intrinsic (entry);
	entry = install_glo ("eoshift_sa", TAB_NAME, 0);
	SET_compile_time (entry);
	SET_no_checking (entry);
	entry->compile_time = intrin_eoshift_sa;
	SET_intrinsic (entry);
	entry = install_glo ("eoshift_aa", TAB_NAME, 0);
	SET_compile_time (entry);
	SET_no_checking (entry);
	entry->compile_time = intrin_eoshift_aa;
	SET_intrinsic (entry);

	entry = install_glo ("project", TAB_NAME, 0);
	SET_intrinsic (entry);
	SET_no_checking (entry);
	entry = install_glo ("project_fs", TAB_NAME, 0);
	SET_compile_time (entry);
	SET_no_checking (entry);
	entry->compile_time = intrin_project_fs;
	SET_intrinsic (entry);
	entry = install_glo ("project_fa", TAB_NAME, 0);
	SET_compile_time (entry);
	SET_no_checking (entry);
	entry->compile_time = intrin_project_fa;
	SET_intrinsic (entry);

	entry = install_glo ("cshift", TAB_NAME, 0);
	SET_intrinsic (entry);
	SET_no_checking (entry);
	entry = install_glo ("cshifta", TAB_NAME, 0);
	SET_compile_time (entry);
	SET_no_checking (entry);
	entry->compile_time = intrin_cshifta;
	SET_intrinsic (entry);
	entry = install_glo ("cshifts", TAB_NAME, 0);
	SET_compile_time (entry);
	SET_no_checking (entry);
	entry->compile_time = intrin_cshifts;
	SET_intrinsic (entry);

	entry = install_glo ("merge", TAB_NAME, 0);
	SET_intrinsic (entry);
	SET_no_checking (entry);

	entry = install_glo ("mergeaaa", TAB_NAME, 0);
	SET_compile_time (entry);
	SET_no_checking (entry);
	entry->compile_time = intrin_mergeaaa;
	SET_intrinsic (entry);

	entry = install_glo ("mergeaas", TAB_NAME, 0);
	SET_compile_time (entry);
	SET_no_checking (entry);
	entry->compile_time = intrin_mergeaas;
	SET_intrinsic (entry);

	entry = install_glo ("mergesss", TAB_NAME, 0);
	SET_compile_time (entry);
	SET_no_checking (entry);
	entry->compile_time = intrin_mergesss;
	SET_intrinsic (entry);
	entry = install_glo ("mergessa", TAB_NAME, 0);
	SET_compile_time (entry);
	SET_no_checking (entry);
	entry->compile_time = intrin_mergessa;
	SET_intrinsic (entry);

	entry = install_glo ("mergeasa", TAB_NAME, 0);
	SET_compile_time (entry);
	SET_no_checking (entry);
	entry->compile_time = intrin_mergeasa;
	SET_intrinsic (entry);

	entry = install_glo ("mergeass", TAB_NAME, 0);
	SET_compile_time (entry);
	SET_no_checking (entry);
	entry->compile_time = intrin_mergeass;
	SET_intrinsic (entry);

	entry = install_glo ("mergesaa", TAB_NAME, 0);
	SET_compile_time (entry);
	SET_no_checking (entry);
	entry->compile_time = intrin_mergesaa;
	SET_intrinsic (entry);

	entry = install_glo ("mergesas", TAB_NAME, 0);
	SET_compile_time (entry);
	SET_no_checking (entry);
	entry->compile_time = intrin_mergesas;
	SET_intrinsic (entry);

	entry = install_glo ("reshape", TAB_NAME, 0);
	SET_compile_time (entry);
	SET_no_checking (entry);
	if (flag_cmf)
		entry->compile_time = intrin_reshape_CMF;
	else
		entry->compile_time = intrin_reshape;
	SET_intrinsic (entry);

	entry = install_glo ("replicate", TAB_NAME, 0);
	SET_compile_time (entry);
	SET_no_checking (entry);
	entry->compile_time = intrin_replicate;
	SET_intrinsic (entry);

	entry = install_glo ("product", TAB_NAME, 0);
	SET_intrinsic (entry);
	SET_no_checking (entry);
	entry = install_glo ("products", TAB_NAME, 0);
	SET_compile_time (entry);
	SET_no_checking (entry);
	entry->compile_time = intrin_products;
	SET_intrinsic (entry);
	entry = install_glo ("producta", TAB_NAME, 0);
	SET_compile_time (entry);
	SET_no_checking (entry);
	entry->compile_time = intrin_producta;
	SET_intrinsic (entry);

	entry = install_glo ("lbound", TAB_NAME, 0);
	SET_no_checking (entry);
	SET_compile_time (entry);
	entry->compile_time = intrin_lbound_F90;
	SET_intrinsic (entry);
	entry = install_glo ("dlbound", TAB_NAME, 0);
	SET_no_checking (entry);
	SET_compile_time (entry);
	entry->compile_time = intrin_lbound_CMF;
	SET_intrinsic (entry);

	entry = install_glo ("ubound", TAB_NAME, 0);
	SET_no_checking (entry);
	SET_compile_time (entry);
	entry->compile_time = intrin_ubound_F90;
	SET_intrinsic (entry);
	entry = install_glo ("dubound", TAB_NAME, 0);
	SET_no_checking (entry);
	SET_compile_time (entry);
	entry->compile_time = intrin_ubound_CMF;
	SET_intrinsic (entry);

#include "intriaut.inc"
}

int intrin_var_args (entry)
T_Entry *entry;
{
	if (!strcmp (entry->name, "MAX")
	||  !strcmp (entry->name, "EEMAX")
	||  !strcmp (entry->name, "AMAX0")
	||  !strcmp (entry->name, "AMAX1")
	||  !strcmp (entry->name, "DMAX1")
	||  !strcmp (entry->name, "MIN")
	||  !strcmp (entry->name, "EEMIN")
	||  !strcmp (entry->name, "AMIN0")
	||  !strcmp (entry->name, "AMIN1")
	||  !strcmp (entry->name, "DMIN1"))
		return 1;
	else
		return 0;
}
